import numpy as np
import scipy
import logging
from typing import Sequence, Tuple, List

# create logging configuration
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# Create a console handler
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)

# Set the formatter for the console handler
formatter = logging.Formatter(
    "%(name)s:%(levelname)s:%(funcName)s:%(message)s",
)
console_handler.setFormatter(formatter)

# Add the console handler to the logger
logger.addHandler(console_handler)


class create_1Dmesh:
    """
    A 1D mesh object.

    Atributes:
    xcell_center (np.array): An array of node x positions
    n_points (int): The number of mesh points
    delta_x (float) : The distance between mesh points

    """

    def __init__(self, x, n_cells, mesh_type="finite_volume"):
        """
        Initialize the Mesh object.

        Keyword Arguments:
        x -- the spatial domain of the mesh in the form [x_min, x_max]
        n_cells -- the number of points for discritization.
        use n_cells as the number of points for the finite volume case

        Example
        mesh = create_1Dmesh(x=[0, 1], n_cells=3)
        mesh.xcell_center = np.array([0.125,0.375, 0.625, 0.875])
        mesh.deltax = 0.25
        """
        self.n_cells = n_cells
        self.mesh_type = mesh_type
        x_grid = grid(n_cells, x, mesh_type)

        self.delta_x = x_grid.cell_width
        self.xcell_center = x_grid.cell_cordinates

        self.x_differentiation_matrix = differentiation_matrix(self.n_cells)
        self.laplacian = self.x_differentiation_matrix.get_matrix()

        self.boundary_condition_object = boundary_condition(
            n_cells=self.n_cells, mesh_type=self.mesh_type
        )
        self.boundary_condition_array = (
            self.boundary_condition_object.boundary_condition_array
        )

        self.cell_phi = cell_phi(n_cells=n_cells, dim=1, mesh_type=mesh_type)


class heat_diffusion_mesh(create_1Dmesh):
    """Create a heat diffusion mesh."""

    def __init__(self, x, n_cells: Sequence[int], mesh_type: str = "finite_volume"):
        """
        Initialize a heat diffusion mesh object.

        Parameters:
           x (type) : the spatial discritization of the domain
           n_cells (int): The number of cells to discritize the domain into
           mesh_type (string) : finite_volume (default) or finite_difference
        """
        super().__init__(x, n_cells, mesh_type)
        self.temperature = self.cell_phi.phi

    def set_cell_temperature(self, temperature):
        """
        Set the temperature for internal nodes.

        Example:running mesh.set_internal_temperature(20) would result
        in np.array([20, 20, 20, 20]
        """
        self.cell_phi.set_phi(phi=temperature)
        self.temperature = self.cell_phi.get_phi()

    def set_thermal_diffusivity(self, thermal_diffusivity):
        """Set a diffusion constant in square meters per second."""
        self.thermal_diffusivity = thermal_diffusivity

    def set_dirichlet_boundary(self, side, temperature):
        """Update boundary array and D2 for a dirichlet boundary."""
        self.x_differentiation_matrix.set_dirichlet_boundary(side, self.mesh_type)
        self.boundary_condition_object.set_dirichlet_boundary(
            side=side, phi=temperature
        )
        self.cell_phi.set_dirichlet_boundary(side, temperature)

    def set_neumann_boundary(self, side, flux=0):
        """Update boundary array and D2 for a neumann boundary."""
        self.x_differentiation_matrix.set_neumann_boundary(side, self.mesh_type)
        self.boundary_condition_object.set_neumann_boundary(
            side=side, flux=flux, cell_width=self.delta_x
        )


class linear_convection_mesh(create_1Dmesh):
    """
    Create a linear convection diffusion mesh.

    Attributes:
    n_cells (int): The number of cells used to discritze the domain
    mesh_type (str): finite_volume or finite_difference
    """

    def __init__(
        self,
        x,
        n_cells: int,
        mesh_type: str = "finite_volume",
        convection_coefficient=1,
        discretization_type: str = "upwind",
    ):
        """
        Initialize a lienar convection mesh object.

        Parameters:
           x : the spatial discritization of the domain
           n_cells (int): The number of cells to discritize the domain into
           mesh_type (string) : finite_voluem (default) or finite_difference
        Attributes:
           phi: the quantity of interest being transported
           convection_coefficent: a constant convection coefficent
        """
        super().__init__(x, n_cells, mesh_type)

        self.discretization_type = discretization_type
        if self.discretization_type == "upwind":
            self.x_differentiation_matrix = upwind_differentiation_matrix(self.n_cells)

        elif self.discretization_type == "central":
            self.x_differentiation_matrix = central_differentiation_matrix(self.n_cells)

        elif self.discretization_type == "maccormack":
            # self.create_maccormack_differentiation_matrix(self.xcell_center)
            self.x_differentiation_matrix = maccormack_differentiation_matrix(
                self.n_cells
            )

            self.predictor_differentiation_matrix = (
                self.x_differentiation_matrix.predictor_differentiation_matrix
            )
        else:
            raise ValueError("discritization type not supported")

        self.laplacian = self.x_differentiation_matrix.get_matrix()

        if convection_coefficient <= 0:
            raise ValueError("only positive convection coefficents are supported")
        self.convection_coefficent = convection_coefficient

        self.phi = cell_phi(n_cells=n_cells, dim=1, mesh_type=mesh_type)

    def set_dirichlet_boundary(self, side: str, phi: float):
        """Update boundary array and D2 for a dirichlet boundary."""
        self.x_differentiation_matrix.set_dirichlet_boundary(side, self.mesh_type)
        # self.boundary_condition_object.set_dirichlet_boundary(side=side, phi=phi / 2)
        self.phi.set_dirichlet_boundary(side, phi)
        side_selector().side_validate(side)
        array_index = side_selector().boundary_index(side)

        if side == "right":
            sign = -1
        elif side == "left":
            sign = 1
        else:
            raise ValueError("oops, you shouldnt be here")
        self.boundary_condition_object.set_dirichlet_boundary(
            side=side, phi=sign * phi / 2
        )

        if self.mesh_type == "finite_volume":
            if self.discretization_type == "central":
                self.laplacian[array_index, array_index] = (-1 / 2) * sign
            elif self.discretization_type == "upwind":
                self.laplacian[array_index, array_index] = -1
            else:
                raise ValueError("oops, unsupported mesh type")

        elif self.mesh_type == "finite_difference":
            if self.discretization_type == "maccormack":
                self.predictor_differentiation_matrix[array_index, :] = 0
        else:
            raise ValueError("mesh must be finite_volume or finite_difference")

    def set_right_boundary(self):
        """
        Set the differentiation matrix using a 2nd order left discritization.

        For finite volume only
        """
        self.laplacian[-1, -3:] = [1.5, -1, 0.5]


class differentiation_matrix:
    """Create a differentiation matrix."""

    def __init__(self, n_cells: int):
        """
        Initialaze a differentiation  matrix.

        Paramaters: n_cells number of cells
        Atributes:
        differentiation_matrix: A sparse  matrix n_cells x n_cells with -2 on the diagonal and a 1 on the +1 and -1 diagonal

        """
        self.__n_cells = n_cells
        self.differentiation_matrix = self.set_diagonal()

    def get_matrix(self):
        """Return: differentiation matrix."""
        return self.differentiation_matrix

    def set_diagonal(self, lower=1, middle=-2, upper=1):
        """Create a sparce diagonal matrix"""
        self.diagonal = np.ones(self.__n_cells)
        return scipy.sparse.spdiags(
            np.array(
                [lower * self.diagonal, middle * self.diagonal, upper * self.diagonal]
            ),
            np.array([-1, 0, 1]),
        ).toarray()

    def set_dirichlet_boundary(self, side, mesh_type):
        """Update boundary array and D2 for a dirichlet boundary."""
        array_index = side_selector().boundary_index(side)

        if mesh_type == "finite_volume":
            self.differentiation_matrix[array_index, array_index] = -3

        elif mesh_type == "finite_difference":
            self.differentiation_matrix[array_index, :] = 0
        else:
            raise ValueError("mesh must be finite_volume or finite_difference")

    def set_neumann_boundary(self, side, mesh_type):
        """Update the differentiation matrix for a neumann boundary."""
        boundary_index = side_selector().boundary_index(side)
        first_interior_index = side_selector().first_interior_index(side)

        if mesh_type == "finite_volume":
            self.differentiation_matrix[boundary_index, boundary_index] = -1
        elif mesh_type == "finite_difference":
            self.differentiation_matrix[boundary_index, first_interior_index] = 2
        else:
            raise ValueError(
                "mesh_type unsupported, please input a finite_volume or finite_difference as mesh type"
            )


class upwind_differentiation_matrix(differentiation_matrix):
    def __init__(self, n_cells: int):
        """Create a differentiation matrix."""
        super().__init__(n_cells)
        self.differentiation_matrix = self.set_diagonal(lower=1, middle=-1, upper=0)


class central_differentiation_matrix(differentiation_matrix):
    def __init__(self, n_cells: int):
        """Create a differentiation matrix."""
        super().__init__(n_cells)
        self.differentiation_matrix = self.set_diagonal(lower=0.5, middle=0, upper=-0.5)


class maccormack_differentiation_matrix(differentiation_matrix):
    def __init__(self, n_cells: int):
        super().__init__(n_cells)
        self.differentiation_matrix = self.set_diagonal(lower=-1, middle=1, upper=0)

        self.predictor_differentiation_matrix = -np.transpose(
            self.differentiation_matrix
        )


class flux_differentiation_matrix:
    pass


class grid:
    """
    A 1d uniformly discritzed grid object.

    Attributes:
        cordinates:
        delta:
    """

    def __init__(
        self, n_cells: int, cordinates: tuple[float, float], mesh_type: str
    ) -> None:
        """
        Args:
            n_cells: int = number of cells to discritize the grid
            cordinates: tupple(float, float) = min and max of the grid
            mesh_type (string)=  finite_volume or finite_difference
        """
        self.n_cells = n_cells
        self.cordinates = cordinates
        self.__discritize(mesh_type)

    def __discritize(self, mesh_type):
        if mesh_type == "finite_volume":
            self.cell_width = (self.cordinates[1] - self.cordinates[0]) / (self.n_cells)
            self.cell_cordinates = np.linspace(
                self.cordinates[0] + (self.cell_width / 2),
                self.cordinates[1] - self.cell_width / 2,
                self.n_cells,
            )
        elif mesh_type == "finite_difference":
            self.cell_width = (self.cordinates[1] - self.cordinates[0]) / (
                self.n_cells - 1
            )
            self.cell_cordinates = np.linspace(
                self.cordinates[0], self.cordinates[1], self.n_cells
            )
        else:
            raise ValueError("Mesh type not supported")


class boundary_condition:
    """
    A boundary condition object.

    Attributes
        boundary_condition_array: array[float]
    """

    def __init__(self, n_cells: int, mesh_type: str):
        self.boundary_condition_array = np.zeros(n_cells)
        mesh_type_validator().validate(mesh_type)
        self.__mesh_type = mesh_type

    def set_dirichlet_boundary(self, side: str, phi: float):
        """Update the boundary contition array for a dirichlet boundary."""
        boundary_index = side_selector().boundary_index(side)
        if self.__mesh_type == "finite_volume":
            self.boundary_condition_array[boundary_index] = 2 * phi
        elif self.__mesh_type == "finite_difference":
            self.boundary_condition_array[boundary_index] = 0

    def set_neumann_boundary(self, side: str, flux: float, cell_width: float):
        """Update the boundary contition array for a neuiman boundary."""
        boundary_index = side_selector().boundary_index(side)
        if self.__mesh_type == "finite_volume":
            # self.boundary_condition_array[boundary_index] = flux / cell_width
            self.boundary_condition_array[boundary_index] = flux * cell_width
        elif self.__mesh_type == "finite_difference":
            self.boundary_condition_array[boundary_index] = 2 * flux * cell_width

    def get_array(self):
        return self.boundary_condition_array


class cell_phi:
    """
    An object that stores each cells phi value.
    """

    def __init__(self, n_cells: list[int] | int, dim: int, mesh_type: str):
        """
        Create phi object and store n_cells.

        args:
        n_cells: List[int] Number of cells, in the order of [n_xcells, n_ycells}
        """
        self.phi = np.zeros(np.flip(n_cells))
        self.__n_cells = n_cells
        mesh_type_validator().validate(mesh_type=mesh_type)
        self.mesh_type = mesh_type
        self.dim = dim

    def set_phi(self, phi):
        """
        Set the value of phi for internal nodes.

        Parameters:
        phi (int, float, list):list of phi at every x value
        """
        if isinstance(phi, (float, int)):
            self.phi.fill(phi)

        elif isinstance(phi, (list)):
            if np.array(phi).shape != self.phi.shape:
                raise ValueError(
                    f"Inputed shape {np.array(phi).shape} does not match phi shape {self.phi.shape} "
                )
            self.phi = np.array(phi)
        else:
            raise TypeError(f"The phi type inputed {type(phi)} not supported")

    def get_phi(self):
        """Return phi object."""
        return self.phi

    def set_dirichlet_boundary(self, side: str, phi: float):
        """Update phi for a dirichlet boundary."""
        boundary_index = side_selector().boundary_index(side)

        if self.mesh_type == "finite_volume":
            pass

        elif self.mesh_type == "finite_difference":
            if self.dim == 2:
                if side in ["left", "right"]:
                    self.phi[:, boundary_index] = phi
                elif side in ["top", "bottom"]:
                    self.phi[boundary_index, :] = phi
            else:
                self.phi[boundary_index] = phi
        else:
            raise ValueError("mesh must be finite_volume or finite_difference")


class side_selector:
    """Set the array index to modify based on the boundary side."""

    def side_validate(self, side: str):
        "Ensure the side is a valid side."
        if side not in ["left", "right", "top", "bottom"]:
            raise ValueError(f"{side} is not left, right, top or bottom")

    def boundary_index(self, side: str):
        """Return the index for the first or last row (or column) based on the side."""
        self.side_validate(side)
        if side == "left" or side == "top":
            return 0
        elif side == "right" or side == "bottom":
            return -1

    def first_interior_index(self, side: str):
        """Return the index for the first interior cell (1, or -2)."""
        self.side_validate(side)
        if side == "left" or side == "top":
            return 1
        elif side == "right" or side == "bottom":
            return -2

    def axis(self, side: str):
        """Return the axis."""
        self.side_validate(side)
        if side == "left" or side == "right":
            return "x"
        elif side == "top" or side == "bottom":
            return "y"


class mesh_type_validator:
    """Validate that the mesh type is supported."""

    def validate(self, mesh_type: str):
        """ensures that the mesh type is either a "finite_volume" or "finite_difference"""
        if mesh_type not in ("finite_volume", "finite_difference"):
            raise ValueError(f"{mesh_type} is not finite_volume or finite_difference")


def main():
    pass


def init():
    if __name__ == "__main__":
        main()


init()



from solver.cartesian_solver import CartesianSolver, SteadySolver
import pytest
import numpy as np


@pytest.fixture
def mock_mesh(mocker):
    """
    Create to create a mock mesh object for use in testing the solver.
    """
    mesh = mocker.MagicMock()
    mesh.laplacian = np.array([[-3, 1, 0], [1, -2, 1], [0, 1, -3]])
    mesh.boundary_condition_array = np.array([60, 0, 0])
    return mesh


class TestCartesianSolver:
    @pytest.fixture
    def lp(self):
        return np.array([[-3, 1, 0], [1, -2, 1], [0, 1, -3]])

    @pytest.fixture
    def bc(self):
        return np.array([60, 0, 0])

    @pytest.fixture
    def solved_temp(self):
        return np.array([25, 15, 5])

    def test_solve_steady(self, lp, bc, solved_temp):
        expectd = solved_temp
        actual = SteadySolver().solve(laplacian=lp, boundary_condition_array=bc)

        np.testing.assert_array_equal(x=actual, y=expectd)

    def test_carteisan_steady_solve(self, lp, bc, solved_temp):
        expectd = solved_temp
        # init = CartesianSolver(method = "steady")
        init = CartesianSolver()
        actual = init.solve(laplacian=lp, boundary_condition_array=bc)
        np.testing.assert_array_equal(x=actual, y=expectd)

    # def test_steady_solver_init(self, lp, bc):
    #     solver = SteadySolver(
    #         laplacian=lp,
    #         boundary_condition_array=bc
    #     )
    #     np.testing.assert_array_equal(x = solver.laplacian, y =  lp)


if __name__ == "__main__":
    pytest.main()



from solver.cartesian_mesh import CartesianMesh
from solver.solver import Solver
from solver.utilities import EnergyBalance, Plotter, VelocityRounder
from solver.mesher import linear_convection_mesh
from matplotlib import pyplot as plt
import numpy as np
import logging
import math

# create logging configuration
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

console_handler = logging.StreamHandler()
console_handler.setLevel(logging.INFO)
formatter = logging.Formatter(
    "%(message)s",
)
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)


class Setup:
    def __init__(
        self,
        x_range=(0, 5),
        y_range=(0, 1),
        mesh_type="finite_volume",
        conductivity=0.456,
        diffusivity=0.146 * 10 ** (-6),
        temp_left=30,
        temp_right=30,
        temp_bottom=30,
        top_flux=-10,
    ):
        self.mesh_size_dict = {
            "course": (5, 2),
            "medium": (20, 8),
            "fine": (100, 10),
            "ultra_fine": (200, 50),
        }
        # Define Problem constants
        self.x_range = x_range
        self.y_range = y_range
        self.mesh_type = mesh_type
        self.conductivity = conductivity  # W/mk
        self.diffusivity = diffusivity  # m^2/s
        self.temp_left = temp_left  # celcius
        self.temp_right = temp_right  # celcius
        self.temp_bottom = temp_bottom  # celcius
        self.top_flux = top_flux  # celcius

        # Create a function that creates the mesh

    def create_mesh(
        self,
        x_cells=80,
        y_cells=12,
    ):
        mesh = CartesianMesh(
            dimensions=2,
            n_cells=[x_cells, y_cells],
            cordinates=[self.x_range, self.y_range],
            mesh_type=self.mesh_type,
            conductivity=self.conductivity,
            diffusivity=self.diffusivity,
        )

        mesh.set_dirichlet_boundary(side="left", phi=self.temp_left)
        mesh.set_dirichlet_boundary(side="right", phi=self.temp_right)
        mesh.set_dirichlet_boundary(side="bottom", phi=self.temp_bottom)
        mesh.set_neumann_boundary(side="top", flux=self.top_flux / self.conductivity)
        return mesh

    def plot_steady(self, mesh_dictionary, plot_name):
        fig, ax = plt.subplots(len(mesh_dictionary), figsize=(10, 12))
        i = 0
        for mesh_size, mesh in mesh_dictionary.items():
            x_cords = mesh.grid["x_grid"].cell_cordinates
            y_cords = mesh.grid["y_grid"].cell_cordinates
            logger.debug(x_cords)
            xv, yv = np.meshgrid(x_cords, y_cords)
            axis = plt.subplot(len(mesh_dictionary), 1, i + 1)
            axis.set_title(mesh_size)
            steady = axis.pcolormesh(xv, yv, mesh.phi.get_phi())

            i = i + 1

            fig.colorbar(steady, label="temperature (celcius)")
        plt.savefig(plot_name)


class SteadyStateSolver:
    def __init__(self, settings=Setup()):
        self.settings = settings

    def solve(self):
        self.mesh_dict = {}
        energy_balance_dict = {}
        for mesh_size, cells in self.settings.mesh_size_dict.items():
            x_cells, y_cells = cells

            self.mesh_dict[mesh_size] = self.settings.create_mesh(
                x_cells=x_cells, y_cells=y_cells
            )
            # Solve the steady Case
            Solver(mesh=self.mesh_dict[mesh_size]).solve_steady()
            energy_balance_dict[mesh_size] = EnergyBalance(
                mesh=self.mesh_dict[mesh_size]
            )
            logger.info(f"\n Mesh Size:{mesh_size}")
            total_flux = energy_balance_dict[mesh_size].flux("all")

    def plot(self):
        fig, ax = plt.subplots(len(self.mesh_dict), figsize=(10, 12))
        i = 0
        for mesh_size, mesh in self.mesh_dict.items():
            x_cords = mesh.grid["x_grid"].cell_cordinates
            y_cords = mesh.grid["y_grid"].cell_cordinates
            logger.debug(x_cords)
            xv, yv = np.meshgrid(x_cords, y_cords)
            axis = plt.subplot(len(self.mesh_dict), 1, i + 1)
            axis.set_title(mesh_size)
            steady = axis.pcolormesh(xv, yv, mesh.phi.get_phi())

            i = i + 1

            fig.colorbar(steady, label="temperature (celcius)")
            plt.savefig("steady.png")


def SteadyState():
    steady = SteadyStateSolver()
    steady.solve()
    steady.plot()


class GenerationSolver:
    def __init__(self, settings=Setup()):
        self.settings = settings

    def solve(self):
        self.generation_mesh_dict = {}
        generation_energy_balance_dict = {}

        def gen_function(x, y):
            return 0 * y + 50 * np.exp(-((x - 2.5) ** 2))

        for mesh_size, cells in self.settings.mesh_size_dict.items():
            x_cells, y_cells = cells

            # Create the mesh
            self.generation_mesh_dict[mesh_size] = self.settings.create_mesh(
                x_cells=x_cells, y_cells=y_cells
            )

            # Add the generation function
            self.generation_mesh_dict[mesh_size].set_generation(function=gen_function)

            # Solve the steady Case with generation
            Solver(mesh=self.generation_mesh_dict[mesh_size]).solve_steady()
            generation_energy_balance_dict[mesh_size] = EnergyBalance(
                mesh=self.generation_mesh_dict[mesh_size]
            )
            logger.info(f"\n Mesh Size:{mesh_size}")
            total_flux = generation_energy_balance_dict[mesh_size].flux("all")

    def plot(self):
        self.settings.plot_steady(self.generation_mesh_dict, "generation.png")


def Generation():
    generation = GenerationSolver()
    generation.solve()
    generation.plot()


class TransientSolver:
    def __init__(self, settings=Setup()):
        self.settings = settings

    def solve(self):
        self.transient_mesh = self.settings.create_mesh()
        self.transient_mesh.phi.set_phi(30)

        # Create the implicit solver
        self.transient_solver = Solver(
            mesh=self.transient_mesh, method="implicit", time_step_size=10000
        )

        self.transient_solver.solve(
            t_final=14000000, record_step=100, compute_error_flag=True, tolerance=0.1
        )

    def plot(self):
        transient_plotter = Plotter(mesh=self.transient_mesh)
        transient_plotter.transient_plotter(
            data_list=self.transient_solver.saved_state_list, name="transient"
        )


def Transient():
    transient = TransientSolver()
    transient.solve()
    transient.plot()


class AdvectionDiffusionMesh:
    def __init__(self, settings=Setup(temp_left=50)):
        self.settings = settings

    def create_advection_diffusion_mesh(self, velocity, x_cells=60, y_cells=20):
        # Create our standard 2d cartesian mesh
        # mesh = self.settings.create_mesh(x_cells = x_cells, y_cells = y_cells, temp_left = 50)
        mesh = self.settings.create_mesh(x_cells=x_cells, y_cells=y_cells)
        mesh.phi.set_phi(30)

        # Create a 1d finite volume linear convective mesh (from HW5)
        convection_mesh = linear_convection_mesh(
            x=self.settings.x_range,
            n_cells=x_cells,
            mesh_type="finite_volume",
            discretization_type="upwind",
        )

        convection_mesh.set_dirichlet_boundary("left", phi=50)

        dx = mesh.grid["x_grid"].cell_width
        Iy = np.identity(y_cells)
        # make the 2d laplacian using the kronecker delta and Iy
        # This puts the 1d laplacian into every 1 in an identity of the shape y_cells x y_cells
        twod_x_lap = (velocity / dx) * np.kron(Iy, convection_mesh.laplacian)
        # logger.debug(f"{twod_x_lap}")
        twod_x_bc = (
            velocity / (dx)
        ) * convection_mesh.boundary_condition_array.reshape(1, x_cells).repeat(
            y_cells, axis=0
        ).flatten()

        # add the convective terms to the 2d diffusion mesh
        mesh.laplacian = mesh.laplacian + twod_x_lap
        mesh.boundary_condition_array = mesh.boundary_condition_array + twod_x_bc
        return mesh


class AdvectionDiffusionSolver:
    def __init__(self, settings=Setup()):
        self.settings = settings

    def solve(self):
        ultra_slow_dict = {
            "name": "ultra_slow",
            "velocity": 0.000001,
            "time_step_size": 10000,
            "t_final": 10000000,
            "record_step_time": 10,
            "tolerance": 0.01,
        }

        mixed_dict = {
            "name": "mixed",
            "velocity": 0.00001,
            "time_step_size": 1000,
            "t_final": 1000000,
            "record_step_time": 10,
            "tolerance": 0.01,
        }

        slow_dict = {
            "name": "slow",
            "velocity": 0.01,
            "time_step_size": 1,
            "t_final": 1000,
            "record_step_time": 10,
            "tolerance": 0.01,
        }

        med_dict = {
            "name": "med",
            "velocity": 0.1,
            "time_step_size": 0.1,
            "t_final": 100,
            "record_step_time": 10,
            "tolerance": 0.01,
        }

        fast_dict = {
            "name": "fast",
            "velocity": 1,
            "time_step_size": 0.01,
            "t_final": 10,
            "record_step_time": 10,
            "tolerance": 0.01,
        }

        settings_list = [fast_dict, med_dict, slow_dict, mixed_dict, ultra_slow_dict]
        # settings_list= [fast_dict]

        # # make a dictionary of solvers for the various time velocities
        self.mesh_dict = {}
        self.solutions_dict = {}
        # Create the mesh
        for settings in settings_list:
            self.mesh_dict[
                settings["name"]
            ] = AdvectionDiffusionMesh().create_advection_diffusion_mesh(
                velocity=settings["velocity"]
            )

            self.solutions_dict[settings["name"]] = Solver(
                mesh=self.mesh_dict[settings["name"]],
                method="implicit",
                time_step_size=settings["time_step_size"],
            )

            self.solutions_dict[settings["name"]].solve(
                t_final=settings["t_final"],
                record_step=settings["record_step_time"],
                compute_error_flag=True,
                tolerance=settings["tolerance"],
            )

    def plot(self):
        for name, solution in self.solutions_dict.items():
            part4_plotter = Plotter(mesh=self.mesh_dict[name]).transient_plotter(
                data_list=solution.saved_state_list,
                name=f"advection_diffusion_{name}",
                phi_min=30,
                phi_max=50,
            )


def AdvectionDiffusion():
    advection_diffusion = AdvectionDiffusionSolver()
    advection_diffusion.solve()
    advection_diffusion.plot()


def main():
    SteadyState()
    Generation()
    Transient()
    AdvectionDiffusion()


if __name__ == "__main__":
    main()



import pytest
import numpy as np
import scipy as sp
from solver.cartesian_mesh import CartesianMesh
from solver.solver import Solver
import logging

logging.basicConfig(format="(levelname)s:%messages)s", level=logging.DEBUG)


@pytest.fixture
def differentation_matrix():
    def _differentiation_matrix(n):
        return sp.sparse.spdiags(
            np.array([np.ones(n), -2 * np.ones(n), np.ones(n)]), np.array([-1, 0, 1])
        ).toarray()

    return _differentiation_matrix


class Test_1d_CartesianMesh:
    @pytest.fixture
    def one_d_mesh(self):
        return CartesianMesh(dimensions=1, cordinates=[(0, 1)], n_cells=[4])

    def test_1d_cell_width(self, one_d_mesh):
        assert one_d_mesh.grid["x_grid"].cell_width == 0.25

    def test_1d_cell_cordinates(self, one_d_mesh):
        np.testing.assert_array_equal(
            x=one_d_mesh.grid["x_grid"].cell_cordinates,
            y=np.array([0.125, 0.375, 0.625, 0.875]),
        )

    def test_1d_differentiation_matrix(self, one_d_mesh, differentation_matrix):
        expected = differentation_matrix(4)
        np.testing.assert_array_equal(
            x=one_d_mesh.differentiation_matrix[
                "x_differentiation_matrix"
            ].get_matrix(),
            y=expected,
        )

    def test_1d_dirichlet_boundary_condition_array(self, one_d_mesh):
        expected = np.array([0, 0, 0, 0])
        np.testing.assert_array_equal(
            x=one_d_mesh.boundary_condition["x_boundary_condition_array"].get_array(),
            y=expected,
        )

    @pytest.mark.parametrize(
        "side,expected",
        [
            (
                "left",
                np.array([[-3, 1, 0, 0], [1, -2, 1, 0], [0, 1, -2, 1], [0, 0, 1, -2]]),
            ),
            (
                "right",
                np.array([[-2, 1, 0, 0], [1, -2, 1, 0], [0, 1, -2, 1], [0, 0, 1, -3]]),
            ),
        ],
    )
    def test_set_dirichlet_boundary_differentiaton_matrix(
        self, one_d_mesh, side, expected
    ):
        one_d_mesh.set_dirichlet_boundary(side=side, phi=30)
        actual = one_d_mesh.differentiation_matrix[
            "x_differentiation_matrix"
        ].get_matrix()
        np.testing.assert_array_equal(x=actual, y=expected)

    @pytest.mark.parametrize(
        "side,expected",
        [
            (
                "left",
                np.array([60, 0, 0, 0]),
            ),
            (
                "right",
                np.array([0, 0, 0, 60]),
            ),
        ],
    )
    def test_set_dirichlet_boundary_boundary_array(self, one_d_mesh, side, expected):
        one_d_mesh.set_dirichlet_boundary(side=side, phi=30)
        actual = one_d_mesh.boundary_condition["x_boundary_condition_array"].get_array()
        np.testing.assert_array_equal(x=actual, y=expected)

    @pytest.mark.parametrize(
        "side,expected",
        [
            (
                "left",
                np.array([30 * 0.25, 0, 0, 0]),
            ),
            (
                "right",
                np.array([0, 0, 0, 30 * 0.25]),
            ),
        ],
    )
    def test_set_neumann_boundary_boundary_array(self, one_d_mesh, side, expected):
        one_d_mesh.set_neumann_boundary(side=side, flux=30)
        actual = one_d_mesh.boundary_condition["x_boundary_condition_array"].get_array()
        np.testing.assert_array_equal(x=actual, y=expected)

    @pytest.mark.parametrize(
        "side,bc_type", [("left", "dirichlet"), ("right", "neumann")]
    )
    def test_save_boundary_conditon(self, one_d_mesh, side, bc_type):
        one_d_mesh.set_dirichlet_boundary(side="left", phi=30),
        one_d_mesh.set_neumann_boundary(side="right", flux=30)
        actual = one_d_mesh.boundary_condition_dict[side]
        assert actual == bc_type

    @pytest.mark.parametrize(
        "side,expected",
        [
            (
                "left",
                np.array([[-1, 1, 0, 0], [1, -2, 1, 0], [0, 1, -2, 1], [0, 0, 1, -2]]),
            ),
            (
                "right",
                np.array([[-2, 1, 0, 0], [1, -2, 1, 0], [0, 1, -2, 1], [0, 0, 1, -1]]),
            ),
        ],
    )
    def test_set_neumann_boundary_differentiaton_matrix(
        self, one_d_mesh, side, expected
    ):
        one_d_mesh.set_neumann_boundary(side=side, flux=30)
        actual = one_d_mesh.differentiation_matrix[
            "x_differentiation_matrix"
        ].get_matrix()
        np.testing.assert_array_equal(x=actual, y=expected)

    # Test a right neuiman, left dirichlet
    def test_laplacian_matrix(self, one_d_mesh):
        one_d_mesh.set_dirichlet_boundary(side="right", phi=30)
        one_d_mesh.set_neumann_boundary(side="left", flux=-10)
        # one_d_mesh.set_laplacian()
        expected = np.array(
            [[-1, 1, 0, 0], [1, -2, 1, 0], [0, 1, -2, 1], [0, 0, 1, -3]]
        ) * (1 / 0.25**2)
        np.testing.assert_array_equal(x=one_d_mesh.laplacian, y=expected)

    # Test a right neuiman, left dirichlet
    def test_boundary_condition_array(self, one_d_mesh):
        one_d_mesh.set_dirichlet_boundary(side="right", phi=30)
        one_d_mesh.set_neumann_boundary(side="left", flux=-10)
        # one_d_mesh.set_laplacian()
        expected = np.array([-10 * 0.25, 0, 0, 60]) / (0.25**2)

        np.testing.assert_array_equal(x=one_d_mesh.boundary_condition_array, y=expected)

    def test_set_phi(self, one_d_mesh):
        one_d_mesh.phi.set_phi(10)
        expected = np.array([10, 10, 10, 10])
        np.testing.assert_array_equal(x=one_d_mesh.phi.get_phi(), y=expected)

    def test_solve(self, one_d_mesh):
        one_d_mesh.set_dirichlet_boundary(side="left", phi=40)
        one_d_mesh.set_dirichlet_boundary(side="right", phi=0)
        expected = np.array([35, 25, 15, 5])
        actual_solver = Solver(mesh=one_d_mesh)
        actual_solver.solve_steady()
        actual = one_d_mesh.phi.get_phi()

        np.testing.assert_array_almost_equal(x=actual, y=expected)

    def test_solve_unsteady(self, one_d_mesh):
        one_d_mesh.set_dirichlet_boundary(side="left", phi=40)
        one_d_mesh.set_dirichlet_boundary(side="right", phi=0)
        expected = np.array([35, 25, 15, 5])
        actual_solver = Solver(mesh=one_d_mesh, method="explicit", time_step_size=0.01)
        actual_solver.solve(10)
        actual = one_d_mesh.phi.get_phi()

        np.testing.assert_array_almost_equal(x=actual, y=expected)

    @pytest.mark.skip
    def test_velocity_differentiation_matrix(self):
        mesh = CartesianMesh(dimensions=1, n_cells=[4], cordinates=[(1, 0)], velocity=5)

        mesh.convective_mesh.differentiation


class Test_2d_CartesianMesh:
    @pytest.fixture
    def two_d_mesh(self):
        return CartesianMesh(dimensions=2, cordinates=[(0, 1), (0, 2)], n_cells=[3, 4])

    cell_width_inputs = [("x_grid", 1 / 3), ("y_grid", 0.5)]

    @pytest.mark.parametrize("dimension,expected", cell_width_inputs)
    def test_2d_cell_width(self, two_d_mesh, dimension, expected):
        assert two_d_mesh.grid[dimension].cell_width == expected

    coordinates_inputs = [
        ("x_grid", np.array([1 / 6, 3 / 6, 5 / 6])),
        ("y_grid", np.array([1.75, 1.25, 0.75, 0.25])),
    ]

    @pytest.mark.parametrize("dimension,expected", coordinates_inputs)
    def test_2d_cell_cordinates(self, two_d_mesh, dimension, expected):
        np.testing.assert_array_equal(
            x=two_d_mesh.grid[dimension].cell_cordinates, y=expected
        )

    differentiation_matrix_inputs = [
        ("x_differentiation_matrix", 3),
        ("y_differentiation_matrix", 4),
    ]

    @pytest.mark.parametrize("dimension,n_cells", differentiation_matrix_inputs)
    def test_2d_cell_differentiation_matrix(
        self, two_d_mesh, dimension, n_cells, differentation_matrix
    ):
        # matrix = getattr(two_d_mesh, dimension)
        expected = differentation_matrix(n_cells)
        np.testing.assert_array_equal(
            x=two_d_mesh.differentiation_matrix[dimension].get_matrix(), y=expected
        )

    boundary_array_inputs = [
        ("x_boundary_condition_array", np.array([0, 0, 0])),
        ("y_boundary_condition_array", np.array([0, 0, 0, 0])),
    ]

    @pytest.mark.parametrize("dimension,expected", boundary_array_inputs)
    def test_2d_cell_boundary_condition_array(
        self, two_d_mesh, dimension, expected, differentation_matrix
    ):
        np.testing.assert_array_equal(
            x=two_d_mesh.boundary_condition[dimension].get_array(), y=expected
        )

    left_dirichlet_diff_matrix = np.array([[-3, 1, 0], [1, -2, 1], [0, 1, -2]])
    right_dirichlet_diff_matrix = np.array([[-2, 1, 0], [1, -2, 1], [0, 1, -3]])
    top_dirichlet_diff_matrix = np.array(
        [[-3, 1, 0, 0], [1, -2, 1, 0], [0, 1, -2, 1], [0, 0, 1, -2]]
    )
    bottom_dirichlet_diff_matrix = np.array(
        [[-2, 1, 0, 0], [1, -2, 1, 0], [0, 1, -2, 1], [0, 0, 1, -3]]
    )

    @pytest.mark.parametrize(
        "name,side,expected",
        [
            ("x_differentiation_matrix", "left", left_dirichlet_diff_matrix),
            ("x_differentiation_matrix", "right", right_dirichlet_diff_matrix),
            ("y_differentiation_matrix", "top", top_dirichlet_diff_matrix),
            ("y_differentiation_matrix", "bottom", bottom_dirichlet_diff_matrix),
        ],
    )
    def test_set_dirichlet_boundary_differentiaton_matrix(
        self, two_d_mesh, name, side, expected
    ):
        two_d_mesh.set_dirichlet_boundary(side=side, phi=30)
        # matrix = getattr(two_d_mesh, name)
        actual = two_d_mesh.differentiation_matrix[name].get_matrix()
        np.testing.assert_array_equal(x=actual, y=expected)

    left_dirichlet_bc_array = np.array([60, 0, 0])
    right_dirichlet_bc_array = np.array([0, 0, 60])
    top_dirichlet_bc_array = np.array([60, 0, 0, 0])
    bottom_dirichlet_bc_array = np.array([0, 0, 0, 60])

    @pytest.mark.parametrize(
        "name,side,expected",
        [
            ("x_boundary_condition_array", "left", left_dirichlet_bc_array),
            ("x_boundary_condition_array", "right", right_dirichlet_bc_array),
            ("y_boundary_condition_array", "top", top_dirichlet_bc_array),
            ("y_boundary_condition_array", "bottom", bottom_dirichlet_bc_array),
        ],
    )
    def test_set_dirichlet_boundary_boundary_array(
        self, two_d_mesh, name, side, expected
    ):
        two_d_mesh.set_dirichlet_boundary(side=side, phi=30)
        actual = two_d_mesh.boundary_condition[name].get_array()
        np.testing.assert_array_equal(x=actual, y=expected)

    left_neumann_diff_matrix = np.array([[-1, 1, 0], [1, -2, 1], [0, 1, -2]])
    right_neumann_diff_matrix = np.array([[-2, 1, 0], [1, -2, 1], [0, 1, -1]])
    top_neumann_diff_matrix = np.array(
        [[-1, 1, 0, 0], [1, -2, 1, 0], [0, 1, -2, 1], [0, 0, 1, -2]]
    )
    bottom_neumann_diff_matrix = np.array(
        [[-2, 1, 0, 0], [1, -2, 1, 0], [0, 1, -2, 1], [0, 0, 1, -1]]
    )

    @pytest.mark.parametrize(
        "name,side,expected",
        [
            ("x_differentiation_matrix", "left", left_neumann_diff_matrix),
            ("x_differentiation_matrix", "right", right_neumann_diff_matrix),
            # ("y_differentiation_matrix", "top", top_neumann_diff_matrix),
            # ("y_differentiation_matrix", "bottom", bottom_neumann_diff_matrix),
        ],
    )
    def test_set_neumann_boundary_differentiaton_matrix(
        self, two_d_mesh, name, side, expected
    ):
        two_d_mesh.set_neumann_boundary(side=side, flux=-10)
        # matrix = getattr(two_d_mesh, name)
        actual = two_d_mesh.differentiation_matrix[name].get_matrix()
        print(side)
        np.testing.assert_array_equal(x=actual, y=expected)

    left_neumann_bc_array = np.array([-10 / 3, 0, 0])
    right_neumann_bc_array = np.array([0, 0, -10 / 3])
    top_neumann_bc_array = np.array([-10 / 2, 0, 0, 0])
    bottom_neumann_bc_array = np.array([0, 0, 0, -10 / 2])

    @pytest.mark.parametrize(
        "name,side,expected",
        [
            ("x_boundary_condition_array", "left", left_neumann_bc_array),
            ("x_boundary_condition_array", "right", right_neumann_bc_array),
            ("y_boundary_condition_array", "top", top_neumann_bc_array),
            ("y_boundary_condition_array", "bottom", bottom_neumann_bc_array),
        ],
    )
    def test_set_neumann_boundary_boundary_array(
        self, two_d_mesh, name, side, expected
    ):
        two_d_mesh.set_neumann_boundary(side=side, flux=-10)
        actual = two_d_mesh.boundary_condition[name].get_array()
        np.testing.assert_array_almost_equal(x=actual, y=expected)

    # Test a mesh that was confirmed to be correct manually
    @pytest.fixture
    def steady_mesh(self):
        steady_mesh = CartesianMesh(
            dimensions=2, cordinates=[(0, 3), (0, 2)], n_cells=[3, 4]
        )
        steady_mesh.set_dirichlet_boundary(side="left", phi=30)
        steady_mesh.set_dirichlet_boundary(side="right", phi=30)
        steady_mesh.set_dirichlet_boundary(side="bottom", phi=30)
        steady_mesh.set_neumann_boundary(side="top", flux=-10)

        return steady_mesh

    def test_laplacian_matrix(self, steady_mesh):
        expected = np.array(
            [
                [-7.0, 1.0, 0.0, 4.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [1.0, -6.0, 1.0, 0.0, 4.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, -7.0, 0.0, 0.0, 4.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [4.0, 0.0, 0.0, -11.0, 1.0, 0.0, 4.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 4.0, 0.0, 1.0, -10.0, 1.0, 0.0, 4.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 4.0, 0.0, 1.0, -11.0, 0.0, 0.0, 4.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 4.0, 0.0, 0.0, -11.0, 1.0, 0.0, 4.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 4.0, 0.0, 1.0, -10.0, 1.0, 0.0, 4.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 4.0, 0.0, 1.0, -11.0, 0.0, 0.0, 4.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 4.0, 0.0, 0.0, -15.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 4.0, 0.0, 1.0, -14.0, 1.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 4.0, 0.0, 1.0, -15.0],
            ]
        )

        np.testing.assert_array_equal(x=steady_mesh.laplacian, y=expected)

    def test_boundary_condition_array(self, steady_mesh):
        expected = np.array(
            [40.0, -20.0, 40.0, 60.0, 0.0, 60.0, 60.0, 0.0, 60.0, 300.0, 240.0, 300.0]
        )
        np.testing.assert_array_equal(
            x=steady_mesh.boundary_condition_array, y=expected
        )

    def test_set_phi(self, two_d_mesh):
        expected = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])
        two_d_mesh.phi.set_phi(expected.tolist())
        np.testing.assert_array_equal(x=two_d_mesh.phi.get_phi(), y=expected)

    @pytest.mark.parametrize(
        "side,bc_type",
        [
            ("left", "dirichlet"),
            ("right", "neumann"),
            ("top", "neumann"),
            ("bottom", "dirichlet"),
        ],
    )
    def test_save_boundary_conditon(self, two_d_mesh, side, bc_type):
        two_d_mesh.set_dirichlet_boundary(side="left", phi=30),
        two_d_mesh.set_neumann_boundary(side="right", flux=30),
        two_d_mesh.set_neumann_boundary(side="top", flux=30)
        two_d_mesh.set_dirichlet_boundary(side="bottom", phi=30)
        actual = two_d_mesh.boundary_condition_dict[side]
        assert actual == bc_type

    def test_conudctivity(self):
        mesh = CartesianMesh(conductivity=10)
        assert mesh.conductivity == 10

    def test_generation_function(self, steady_mesh):
        """test that the function f = x modifies the bc array"""

        expected = np.array(
            [40.5, -18.5, 42.5, 60.5, 1.5, 62.5, 60.5, 1.5, 62.5, 300.5, 241.5, 302.5]
        )
        steady_mesh.set_dirichlet_boundary

        def gen_function(x, y):
            return 0 * y + x

        steady_mesh.set_generation(function=gen_function)
        np.testing.assert_array_equal(
            x=steady_mesh.boundary_condition_array, y=expected
        )

    @pytest.mark.xfail(reason="finite difference not implemented")
    def test_finite_diff_mesh(self):
        fd_mesh = CartesianMesh(
            dimensions=2,
            cordinates=[(0, 3), (0, 2)],
            n_cells=[3, 4],
            mesh_type="finite_difference",
        )
        fd_mesh.set_dirichlet_boundary(side="left", phi=10)
        fd_mesh.set_dirichlet_boundary(side="right", phi=20)
        fd_mesh.set_dirichlet_boundary(side="bottom", phi=30)
        fd_mesh.set_dirichlet_boundary(side="top", phi=40)

        expected = np.array([[40, 40, 40], [10, 0, 20], [10, 0, 20], [30, 30, 30]])
        np.testing.assert_array_equal(x=fd_mesh.phi.get_phi(), y=expected)


class Test_CartesianMesh_exceptions:
    """Test features expected to raise an exception"""

    exception_inputs = [
        ({"dimensions": 3}),
        ({"dimensions": 2, "cordinates": [(0, 1)]}),
        ({"dimensions": 2, "n_cells": [5]}),
        ({"dimensions": 1}),
        # ({"mesh_type": "finite_difference"}),
    ]

    @pytest.mark.parametrize("inputs", exception_inputs)
    def test_three_dimensions_raises(self, inputs):
        with pytest.raises(ValueError):
            CartesianMesh(**inputs)

    def test_side_selector(self):
        with pytest.raises(ValueError):
            CartesianMesh().set_dirichlet_boundary(side="front", phi=10)



import numpy as np
import pandas as pd
from typing import List
from solver.cartesian_mesh import CartesianMesh
import logging

# create logging configuration
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# Create a console handler
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.INFO)

# Set the formatter for the console handler
formatter = logging.Formatter(
    "%(name)s:%(levelname)s:%(funcName)s:%(message)s",
)
console_handler.setFormatter(formatter)

# Add the console handler to the logger
logger.addHandler(console_handler)


class ImplicitStep(object):
    """An object to take an implicit step"""

    def step(self, k, laplacian, boundary_condition_array, phi):
        """Solve the form ay = x + b.

        y = returned
        a = [k*laplacian + I]
        x = phi
        b = k + boundary_condition_array
        """
        # solve the form ay = x+b where x= current temp, y= new temp
        identity_matrix = np.identity(laplacian.shape[0])
        a = -k * laplacian + identity_matrix
        b = k * boundary_condition_array

        return np.linalg.solve(a, (phi + b))


class ExplicitStep(object):
    """An object to take an explicit step."""

    def step(self, k, laplacian, boundary_condition_array, phi):
        """Solve the form y = ax + b.

        y = returned
        a = [k*laplacian + I]
        x = phi
        b = k + boundary_condition_array
        """
        identity_matrix = np.identity(laplacian.shape[0])
        a = k * laplacian + identity_matrix
        b = k * boundary_condition_array
        return a @ phi + b


class SteadySolver(object):
    def solve(self, laplacian, boundary_condition_array):
        return np.linalg.solve(laplacian, -boundary_condition_array)


class Stepper(object):
    """A stepper object to determine the step type to take."""

    def __init__(self, ExplicitStep=ExplicitStep(), ImplicitStep=ImplicitStep()):
        self.explicit_step = ExplicitStep
        self.implicit_step = ImplicitStep

    def take_step(self, method, **kwags):
        """Take the appropriate step depending on the method."""
        if method == "explicit":
            return self.explicit_step.step(**kwags)
        if method == "implicit":
            return self.implicit_step.step(**kwags)


class Saver(object):
    """An object that saves the state of the solution to a data frame."""

    def __init__(self):
        self.saved_state_list = []

    def save_state(self, record_type="data_frame", *args, **kwargs):
        """
        Save the object atributes specified by appending it to the saved_state_list.

        Inputs
        *args must be atributes of the object

        Outputs:
        Appends to a self.saved_state_list
        """
        saved_state_dictionary = {}
        for arg in args:
            saved_state_dictionary[arg] = getattr(self, arg)
        for key, value in kwargs.items():
            saved_state_dictionary[key] = value
        if record_type == "dictionary":
            self.saved_state_list.append(saved_state_dictionary)
        else:
            self.saved_state_list.append(pd.DataFrame(saved_state_dictionary))


class Solver(Saver):
    """main solver class."""

    def __init__(
        self,
        mesh,
        initial_time=0,
        time_step_size=1,
        method="explicit",
        stepper=Stepper(),
        steady_solver=SteadySolver(),
    ):
        """Initiate the main solver:

        args:
        mesh: mesh object (reqires laplacian and boundary_condition_array atributes)
        initial_time:
        """
        self.initial_time = initial_time
        self.step_size = time_step_size
        self.method = method

        self.saved_state_list = []
        self.current_time = initial_time
        self.mesh = mesh
        self.laplacian = self.mesh.laplacian
        self.boundary_condition_array = self.mesh.boundary_condition_array
        self.stepper = stepper
        self.steady_solver = steady_solver

    def take_step(self, k, atribute):
        return self.stepper.take_step(
            method=self.method,
            k=k,
            laplacian=self.laplacian,
            boundary_condition_array=self.boundary_condition_array,
            phi=atribute,
        )

    def solve_steady(self):
        """Call the steady solver"""
        phi_shape = self.mesh.phi.get_phi().shape

        solved_phi = self.steady_solver.solve(
            laplacian=self.laplacian,
            boundary_condition_array=self.boundary_condition_array,
        )
        phi_reshape = np.reshape(solved_phi, phi_shape)
        self.mesh.phi.set_phi(phi_reshape.tolist())

    def solve(
        self, t_final, t_initial=0, record_step=1, compute_error_flag=False, tolerance=0
    ):
        """
        Run the solver for unitil the final time is reached.

        Inputs:
        t_initial = the initial time (default 0)
        t_final = the final time
        record_step = how often to record the solution
        compute_error: boolian = should the steady state solution be calcualted to show how far off steady the solution is

        """
        self.compute_error_flag = compute_error_flag

        self.update_save_dictionary(phi=self.mesh.phi.get_phi())
        super().save_state(record_type="dictionary", **self.save_dictionary)

        self.current_time = t_initial
        # self.error = None
        time_save_index = 1
        while self.current_time < t_final:
            phi = self.mesh.phi.get_phi()
            phi.shape = phi.shape
            solved_phi = self.take_step(k=self.step_size, atribute=phi.flatten())

            phi_reshape = np.reshape(solved_phi, phi.shape)
            self.mesh.phi.set_phi(phi_reshape.tolist())

            self.current_time = self.current_time + self.step_size
            if time_save_index == record_step:
                self.update_save_dictionary(phi=self.mesh.phi.get_phi())
                super().save_state(record_type="dictionary", **self.save_dictionary)

                error = self.compute_error(phi=solved_phi)
                if (not error == None) and (error < tolerance):
                    logger.info(
                        f"steady state reached with an error: {error} < tolerance: {tolerance}"
                    )
                    break
                time_save_index = 0

            time_save_index = time_save_index + 1

        if not time_save_index == 1:
            super().save_state(record_type="dictionary", **self.save_dictionary)
        self.error = self.compute_error(phi=solved_phi)

    def compute_error(self, phi):
        if self.compute_error_flag:
            self.store_steady()
            percent_difference = (phi - self.steady_phi) / self.steady_phi
            rmse = np.sqrt(np.sum(percent_difference**2))
            logger.info(f"Time: {self.current_time}, RMSPE: {rmse}")
            return rmse
        else:
            return None

    def store_steady(self):
        # Calculate and store the steady state phi if it does not exist
        if not hasattr(self, "steady_phi"):
            self.steady_phi = self.steady_solver.solve(
                laplacian=self.laplacian,
                boundary_condition_array=self.boundary_condition_array,
            )

    def update_save_dictionary(self, **kwargs):
        self.save_dictionary = {
            "method": self.method,
            "time_step_size": self.step_size,
            "time": self.current_time,
        }
        for key, value in kwargs.items():
            self.save_dictionary[key] = value


class solver_1d(Solver):
    def __init__(self, mesh, initial_time=0, time_step_size=1, method="explicit"):
        super().__init__(mesh, initial_time, time_step_size, method)
        self.time_step_size = self.step_size

    def take_step(self):
        """
        Take a single step forward in temprature.

        Inputs:
        delta_t: the time step size to take
        """
        k = (
            self.mesh.thermal_diffusivity
            * self.time_step_size
            / (self.mesh.delta_x**2)
        )
        self.mesh.temperature = super().take_step(k, self.mesh.temperature)

    def solve(self, t_final, t_initial=0):
        """
        Run the solver for unitil the final time is reached.

        Inputs:
        t_initial = the initial time (default 0)
        t_final = the final time
        """

        self.current_time = t_initial
        super().update_save_dictionary(
            x_cordinates=self.mesh.xcell_center, temperature=self.mesh.temperature
        )
        super().save_state(**self.save_dictionary)
        while self.current_time < t_final:
            self.take_step()
            self.current_time = self.current_time + self.time_step_size
            self.update_save_dictionary(
                temperature=self.mesh.temperature, x_cordinates=self.mesh.xcell_center
            )
            super().save_state(**self.save_dictionary)

        # # Save the data into a single data frame for ploting
        self.saved_data = pd.concat(self.saved_state_list)


class linear_convection_solver(Solver):
    def __init__(self, mesh, initial_time=0, time_step_size=1, method="explicit"):
        super().__init__(mesh, initial_time, time_step_size, method)
        self.predictor_laplacian = mesh.preictor_laplacian
        self.time_step_size = self.step_size

    def take_step(self):
        """
        Take a single step forward in time.

        Inputs:
        delta_t: the time step size to take
        """
        k = self.mesh.convection_coefficent * self.time_step_size / (self.mesh.delta_x)
        self.courant_coefficent = k
        if self.mesh.discretization_type == "maccormack":
            self.mesh.phi = self.maccormack_take_step(k, self.mesh.phi)
        else:
            self.mesh.phi = super().take_step(k, self.mesh.phi)

    def maccormack_take_step(self, k, atribute):
        laplacian = self.mesh.laplacian
        identity_matrix = np.identity(self.mesh.n_cells)
        predictor_matrix = self.mesh.predictor_differentiation_matrix
        # self.predictor = (identity_matrix - predictor_matrix)@atribute
        if self.method == "explicit":
            self.predictor = (identity_matrix - k * predictor_matrix) @ atribute

            return 0.5 * (
                atribute + ((identity_matrix - k * laplacian) @ self.predictor)
            )

        elif self.method == "implicit":
            raise Exception("implicit not implemented for maccormack")
        else:
            raise Exception("implicit or explicit method needed")

    def solve(self, t_final, t_initial=0):
        """
        Run the solver for unitil the final time is reached.

        Inputs:
        t_initial = the initial time (default 0)
        t_final = the final time
        """

        self.current_time = t_initial
        self.courant_coefficent = (
            self.mesh.convection_coefficent * self.time_step_size / (self.mesh.delta_x)
        )

        self.update_save_dictionary(
            phi=self.mesh.phi,
            courant=self.courant_coefficent,
            discritization=self.mesh.discretization_type,
        )

        self.save_state(**self.save_dictionary)
        while self.current_time < t_final:
            self.take_step()
            self.current_time = self.current_time + self.time_step_size

            self.update_save_dictionary(
                phi=self.mesh.phi,
                courant=self.courant_coefficent,
                discritization=self.mesh.discretization_type,
            )
            self.save_state(**self.save_dictionary)

        # # Save the data into a single data frame for ploting
        self.saved_data = pd.concat(self.saved_state_list)


def main():
    pass


def init():
    if __name__ == "__main__":
        main()


# init()



import numpy as np
import pandas as pd
import pytest
from solver.solver import solver_1d, SteadySolver, Solver
from solver.mesher import heat_diffusion_mesh
from solver import solver
from solver import cartesian_mesh

from unittest.mock import patch, MagicMock


# TODO move this integration test to its own section
# Create a mesh for some integration testing with the meshing
@pytest.fixture
def four_cell_mesh():
    mesh = heat_diffusion_mesh(x=[0, 1], n_cells=4)
    mesh.set_cell_temperature(0)
    mesh.set_dirichlet_boundary("left", 50)
    mesh.set_neumann_boundary("right")
    mesh.set_thermal_diffusivity(0.0001)
    return mesh


@pytest.fixture
def integration_test_explicit_solver(four_cell_mesh):
    return solver_1d(
        mesh=four_cell_mesh,
        initial_time=0,
        time_step_size=1,
        method="explicit",
    )


def test_initiate_solver(integration_test_explicit_solver):
    assert integration_test_explicit_solver.initial_time == 0
    assert integration_test_explicit_solver.time_step_size == 1
    assert integration_test_explicit_solver.method == "explicit"
    assert integration_test_explicit_solver.mesh.n_cells == 4
    assert integration_test_explicit_solver.mesh.thermal_diffusivity == 0.0001


## End integration test


@pytest.fixture
def mock_mesh(mocker):
    """
    Create to create a mock mesh object for use in testing the solver.
    Mesh configuration
    N_elements = 4
    Left boundary = Dirchilet at 50c
    Right boundary = neuiman with q=0
    :return: mesh
    """
    mesh = mocker.MagicMock()
    mesh.xcell_center = np.array([0.125, 0.375, 0.625, 0.875])
    mesh.delta_x = 0.25
    mesh.temperature = np.array([0, 0, 0, 0])
    mesh.n_cells = 4
    mesh.thermal_diffusivity = 0.0001
    mesh.laplacian = np.array(
        [[-3, 1, 0, 0], [1, -2, 1, 0], [0, 1, -2, 1], [0, 0, 1, -1]]
    )
    mesh.boundary_condition_array = np.array(np.array([100, 0, 0, 0]))

    return mesh


@pytest.fixture
def explicit_solver(mock_mesh):
    return solver_1d(
        mesh=mock_mesh,
        initial_time=0,
        time_step_size=1,
        method="explicit",
    )


# # Create test cases to test the implicit solver
@pytest.fixture
def implicit_solver(mock_mesh):
    return solver_1d(
        mesh=mock_mesh,
        initial_time=0,
        time_step_size=1,
        method="implicit",
    )


@pytest.mark.parametrize(
    "solver_fixture, expected_method",
    [
        ("explicit_solver", "explicit"),
        ("integration_test_explicit_solver", "explicit"),
        ("implicit_solver", "implicit"),
    ],
)
def test_solver_initiation(solver_fixture, expected_method, request):
    solver_instance = request.getfixturevalue(solver_fixture)
    assert solver_instance.initial_time == 0
    assert solver_instance.time_step_size == 1
    assert solver_instance.method == expected_method
    assert solver_instance.mesh.n_cells == 4
    assert solver_instance.mesh.thermal_diffusivity == 0.0001


# Expected temperature after steping forward 1 and 3 steps
step_forward_expected_results = [
    ("explicit_solver", np.array([0.16, 0, 0, 0]), np.array([0.4777, 0.000766, 0, 0])),
    (
        "integration_test_explicit_solver",
        np.array([0.16, 0, 0, 0]),
        np.array([0.4777, 0.000766, 0, 0]),
    ),
    (
        "implicit_solver",
        np.array([1.59236073e-01, 2.53965675e-04, 4.05049955e-07, 6.47044657e-10]),
        np.array([4.75432619e-01, 1.51572460e-03, 4.02797229e-06, 9.65628626e-09]),
    ),
]


@pytest.mark.parametrize(
    "solver_fixture, expected, three_step_expected", step_forward_expected_results
)
def test_solver_take_step_new(solver_fixture, expected, three_step_expected, request):
    solver_instance = request.getfixturevalue(solver_fixture)
    solver_instance.take_step()
    expected_temperature = expected
    np.testing.assert_almost_equal(
        solver_instance.mesh.temperature, expected_temperature, decimal=5
    )


# Test that the solver can be called
@pytest.mark.parametrize(
    "solver_fixture, expected, three_step_expected", step_forward_expected_results
)
def test_solver_solve(solver_fixture, expected, three_step_expected, request):
    solver_instance = request.getfixturevalue(solver_fixture)
    solver_instance.solve(t_initial=0, t_final=1)
    expected_temperature = expected
    np.testing.assert_almost_equal(
        solver_instance.mesh.temperature, expected_temperature, decimal=5
    )


@pytest.mark.parametrize(
    "solver_fixture, expected, three_step_expected", step_forward_expected_results
)
def test_solver_solve_multiple_timesteps(
    solver_fixture, expected, three_step_expected, request
):
    solver_instance = request.getfixturevalue(solver_fixture)
    solver_instance.solve(t_initial=0, t_final=3)
    expected_temperature = three_step_expected
    np.testing.assert_almost_equal(
        solver_instance.mesh.temperature, expected_temperature, decimal=5
    )


def test_solver_save_creates_saved_state(explicit_solver):
    solver_instance = explicit_solver
    solver_instance.save_state()
    assert hasattr(solver_instance, "saved_state_list")


@pytest.mark.xfail(reason="pandas dataframe unable to determine index")
def test_solver_save_state_accepts_atribute_names(explicit_solver):
    solver_instance = explicit_solver
    solver_instance.save_state("time_step_size")
    expected_list = pd.concat([pd.DataFrame({"time_step_size": 1}, index=[0])])
    pd.testing.assert_frame_equal(
        pd.concat(solver_instance.saved_state_list), expected_list
    )


def test_solver_save_state_accepts_keywords(explicit_solver):
    solver_instance = explicit_solver
    solver_instance.save_state(
        method="explicit", x_position=np.array([0.125, 0.375, 0.625, 0.875])
    )
    expected_list = pd.concat(
        [
            pd.DataFrame(
                {
                    "method": "explicit",
                    "x_position": np.array([0.125, 0.375, 0.625, 0.875]),
                }
            )
        ]
    )
    pd.testing.assert_frame_equal(
        pd.concat(solver_instance.saved_state_list), expected_list
    )


# @pytest.mark.xfail(reason="data type of time_step_size int64 vs float64")
def test_integration_solve_save_state(explicit_solver):
    solver_instance = explicit_solver
    solver_instance.solve(t_initial=0, t_final=1)
    time_zero_dict = dict(
        method="explicit",
        time_step_size=1,
        time=0,
        x_cordinates=np.array([0.125, 0.375, 0.625, 0.875]),
        temperature=np.array([0, 0, 0, 0]),
    )
    time_one_dict = dict(
        method="explicit",
        time_step_size=1,
        time=1,
        x_cordinates=np.array([0.125, 0.375, 0.625, 0.875]),
        temperature=np.array([0.16, 0, 0, 0]),
    )
    expected_list = []
    expected_list.append(pd.DataFrame(time_zero_dict))
    expected_list.append(pd.DataFrame(time_one_dict))
    expected_data_frame = pd.concat(expected_list)

    pd.testing.assert_frame_equal(
        pd.concat(solver_instance.saved_state_list), expected_data_frame
    )

    pd.testing.assert_frame_equal(solver_instance.saved_data, expected_data_frame)


def test_init():
    from solver import solver

    with patch.object(solver, "main", MagicMock()) as mock_main:
        with patch.object(solver, "__name__", "__main__"):
            solver.init()
    mock_main.assert_called_once()


# from unittest import mock
# def test_init():
#     from solver import solver


#     with mock.patch.object(solver, "main", return_value = 42):
#         with mock.patch.object(solver, "__name__", "__main__"):
#             with mock.patch.object(solver.sys,'exit') as mock_exit:
#                 solver.init()
#                 assert mock_exit.call_args[0][0] == 42
@pytest.fixture
def mock_linear_convective_mesh_upwind(mocker):
    """
    Create to create a mock mesh object for use in testing the solver.
    Mesh configuration
    N_elements = 4
    Left boundary = Dirchilet 1
    :return: mesh
    """
    mesh = mocker.MagicMock()
    mesh.xcell_center = np.array([0.125, 0.375, 0.625, 0.875])
    mesh.delta_x = 0.25
    mesh.phi = np.array([1, 1, 0, 0])
    mesh.n_cells = 4
    mesh.laplacian = np.array(
        [[-1, 0, 0, 0], [1, -1, 0, 0], [0, 1, -1, 0], [0, 0, 1, -1]]
    )
    mesh.boundary_condition_array = np.array(np.array([1, 0, 0, 0]))
    mesh.convection_coefficent = 1

    mesh.discretization_type = "upwind"
    return mesh


@pytest.fixture
def upwind_solver(mock_linear_convective_mesh_upwind):
    return solver.linear_convection_solver(
        mesh=mock_linear_convective_mesh_upwind,
        initial_time=0,
        time_step_size=0.25,
        method="explicit",
    )


def test_upwind_take_step_(upwind_solver):
    upwind_solver.take_step()
    expected_phi = [1, 1, 1, 0]

    np.testing.assert_array_equal(upwind_solver.mesh.phi, expected_phi)


def test_upwind_solver(upwind_solver):
    upwind_solver.solve(2)
    expected_phi = [1, 1, 1, 1]

    np.testing.assert_array_equal(upwind_solver.mesh.phi, expected_phi)


@pytest.fixture
def mock_linear_convective_mesh_mcormack(mocker):
    """
    Create to create a mock mesh object for use in testing the solver.
    Mesh configuration
    N_elements = 4
    Left boundary = Dirchilet 1
    :return: mesh
    """
    mesh = mocker.MagicMock()
    mesh.xcell_center = np.array([0, 0.25, 0.5, 0.75, 1])
    mesh.delta_x = 0.25
    mesh.phi = np.array([1, 1, 0, 0, 0])
    mesh.n_cells = 5
    mesh.laplacian = np.array(
        [
            [0, 0, 0, 0, 0],
            [-1, 1, 0, 0, 0],
            [0, -1, 1, 0, 0],
            [0, 0, -1, 1, 0],
            [0, 0, 0, -1, 1],
        ]
    )
    mesh.predictor_differentiation_matrix = np.array(
        [
            [0, 0, 0, 0, 0],
            [0, -1, 1, 0, 0],
            [0, 0, -1, 1, 0],
            [0, 0, 0, -1, 1],
            [0, 0, 0, 0, -1],
        ]
    )
    mesh.boundary_condition_array = np.array(np.array([0, 0, 0, 0, 0]))
    mesh.convection_coefficent = 1

    mesh.discretization_type = "maccormack"
    return mesh


@pytest.fixture
def maccormack_solver(mock_linear_convective_mesh_mcormack):
    return solver.linear_convection_solver(
        mesh=mock_linear_convective_mesh_mcormack,
        initial_time=0,
        time_step_size=0.25,
        method="explicit",
    )


def test_maccormack_take_step_predictor(maccormack_solver):
    maccormack_solver.take_step()
    expected_predictor = [1, 2, 0, 0, 0]

    np.testing.assert_array_equal(maccormack_solver.predictor, expected_predictor)


def test_maccormack_take_step(maccormack_solver):
    maccormack_solver.take_step()
    expected_phi = [1, 1, 1, 0, 0]

    np.testing.assert_array_equal(maccormack_solver.mesh.phi, expected_phi)


@pytest.fixture
def solved_temp_1d():
    return np.array([25, 15, 5])


class TestSteadySolver:
    @pytest.fixture
    def lp(self):
        return np.array([[-3, 1, 0], [1, -2, 1], [0, 1, -3]])

    @pytest.fixture
    def bc(self):
        return np.array([60, 0, 0])

    # @pytest.fixture
    # def solved_temp(self):
    #     return np.array([25, 15, 5])

    def test_solve_steady(self, lp, bc, solved_temp_1d):
        expectd = solved_temp_1d
        actual = SteadySolver().solve(laplacian=lp, boundary_condition_array=bc)
        np.testing.assert_array_equal(x=actual, y=expectd)


class TestCartesianMesh_Integration:
    @pytest.fixture
    def CartesianMesh_1d(self):
        mesh = cartesian_mesh.CartesianMesh(
            dimensions=1, n_cells=[3], cordinates=[(0, 1)]
        )
        mesh.set_dirichlet_boundary("left", 30)
        mesh.set_dirichlet_boundary("right", 0)
        return mesh

    def test_solve_steady(
        self, CartesianMesh_1d, solved_temp_1d
    ):  # CartesianMesh_1d, solved_temp_1d):
        actual = solver.Solver(mesh=CartesianMesh_1d)
        actual.solve_steady()
        actual_phi = actual.mesh.phi.get_phi()
        expected = solved_temp_1d
        np.testing.assert_array_equal(x=actual_phi, y=expected)

    @pytest.fixture
    def CartesianMesh_2d(self):
        mesh = cartesian_mesh.CartesianMesh(
            dimensions=2, n_cells=[3, 4], cordinates=[(0, 1), (0, 2)]
        )
        mesh.set_dirichlet_boundary("left", 30)
        mesh.set_dirichlet_boundary("right", 30)
        mesh.set_dirichlet_boundary("bottom", 30)
        mesh.set_neumann_boundary("top", -10)
        return mesh

    @pytest.fixture
    def steady_2d_solved(self):
        return np.array(
            [
                [28.72076498, 27.94615599, 28.72076498],
                [29.70707761, 29.46041556, 29.70707761],
                [29.93022914, 29.86469587, 29.93022914],
                [29.98686165, 29.97407644, 29.98686165],
            ]
        )

    def test_solve_steady_2d(self, CartesianMesh_2d, steady_2d_solved):
        actual = solver.Solver(mesh=CartesianMesh_2d)
        actual.solve_steady()
        actual_phi = actual.mesh.phi.get_phi()
        expected = steady_2d_solved
        np.testing.assert_array_almost_equal(x=actual_phi, y=expected)

    def test_solve_unsteady_2d(self, CartesianMesh_2d, steady_2d_solved):
        actual = solver.Solver(
            mesh=CartesianMesh_2d,
            method="implicit",
        )
        actual.solve(t_final=20)
        actual_phi = actual.mesh.phi.get_phi()
        expected = steady_2d_solved
        # verify it matches the steady case
        np.testing.assert_array_almost_equal(x=actual_phi, y=expected)

    def test_unsteady_error_report_2d(self, CartesianMesh_2d, steady_2d_solved):
        """Test that the solver can report the difference between itself and steady state"""
        actual = solver.Solver(
            mesh=CartesianMesh_2d,
            method="implicit",
        )
        actual.solve(t_final=20, compute_error_flag=True)
        actual_phi = actual.mesh.phi.get_phi()
        steady_phi = steady_2d_solved

        expected = np.sqrt(np.sum((actual_phi - steady_phi) ** 2 / steady_phi))
        np.testing.assert_almost_equal(actual.error, expected)

    def test_record_time_2d(self, CartesianMesh_2d, steady_2d_solved):
        """Test that the solver can take an input to record every x timesteps"""
        actual = solver.Solver(
            mesh=CartesianMesh_2d,
            method="implicit",
        )
        # record_list = list()
        actual.solve(t_final=20, compute_error_flag=True, record_step=10)

        # verify it matches the steady case
        assert len(actual.saved_state_list) == 3

        actual_phi = actual.saved_state_list[2]["phi"]
        expected = steady_2d_solved
        # verify it matches the steady case
        np.testing.assert_array_almost_equal(x=actual_phi, y=expected)


if __name__ == "__main__":
    pytest.main()



import pytest
import numpy as np
from solver.mesher import heat_diffusion_mesh
from solver.mesher import create_1Dmesh
from unittest.mock import patch, MagicMock
from solver import mesher


class Test_mesh:
    n_cells: int = 4
    x_range = [0, 1]
    mesh_type: str = "finite_volume"
    expected_xcell_center = np.array([0.125, 0.375, 0.625, 0.875])
    expected_delta_x: float = 0.25

    @pytest.fixture
    def mesh_fixture(self):
        return create_1Dmesh(
            self.x_range, n_cells=self.n_cells, mesh_type=self.mesh_type
        )

    def test_set_n_cells(self, mesh_fixture):
        assert mesh_fixture.n_cells == self.n_cells

    def test_discritization(self, mesh_fixture):
        np.testing.assert_allclose(
            actual=mesh_fixture.xcell_center,
            desired=self.expected_xcell_center,
            atol=0.00001,
        )

    def test_delta_x(self, mesh_fixture):
        assert mesh_fixture.delta_x == self.expected_delta_x


############################################################
class Test_maccormac_linear_convection_mesh(Test_mesh):
    n_cells = 5
    mesh_type = "finite_difference"

    # Expected Paramaters
    expected_xcell_center = np.array([0, 0.25, 0.5, 0.75, 1])
    expected_n_cells = 5
    expected_laplacian = np.array(
        [
            [1, 0, 0, 0, 0],
            [-1, 1, 0, 0, 0],
            [0, -1, 1, 0, 0],
            [0, 0, -1, 1, 0],
            [0, 0, 0, -1, 1],
        ]
    )

    expected_predictor_differentiation_matrix = np.array(
        [
            [-1, 1, 0, 0, 0],
            [0, -1, 1, 0, 0],
            [0, 0, -1, 1, 0],
            [0, 0, 0, -1, 1],
            [0, 0, 0, 0, -1],
        ]
    )

    expected_left_dirichlet_laplacian = np.array(
        [
            [0, 0, 0, 0, 0],
            [-1, 1, 0, 0, 0],
            [0, -1, 1, 0, 0],
            [0, 0, -1, 1, 0],
            [0, 0, 0, -1, 1],
        ]
    )
    expected_left_dirichlet_predctor_laplacian = np.array(
        [
            [0, 0, 0, 0, 0],
            [0, -1, 1, 0, 0],
            [0, 0, -1, 1, 0],
            [0, 0, 0, -1, 1],
            [0, 0, 0, 0, -1],
        ]
    )

    expected_left_dirichlet_boundary_condition_array = np.array([0, 0, 0, 0, 0])

    expected_left_dirichlet_phi = np.array([5, 0, 0, 0, 0])

    @pytest.fixture
    def mesh_fixture(self):
        return mesher.linear_convection_mesh(
            self.x_range,
            n_cells=self.n_cells,
            mesh_type=self.mesh_type,
            convection_coefficient=1,
            discretization_type="maccormack",
        )

    def test_maccormack_laplacian(self, mesh_fixture):
        np.testing.assert_array_equal(mesh_fixture.laplacian, self.expected_laplacian)

    def test_maccormack_predictor_differentiation_matrix(self, mesh_fixture):
        np.testing.assert_array_equal(
            mesh_fixture.predictor_differentiation_matrix,
            self.expected_predictor_differentiation_matrix,
        )

    def test_mcormak_left_dirichlet_laplacian(self, mesh_fixture):
        mesh_fixture.set_dirichlet_boundary("left", 5)
        np.testing.assert_array_equal(
            mesh_fixture.laplacian,
            self.expected_left_dirichlet_laplacian,
        )

    def test_mcormak_left_dirichlet_predictor_differentiation_matrix(
        self, mesh_fixture
    ):
        mesh_fixture.set_dirichlet_boundary("left", 5)
        np.testing.assert_array_equal(
            mesh_fixture.predictor_differentiation_matrix,
            self.expected_left_dirichlet_predctor_laplacian,
        )


############################################################
class Test_central_linear_convection_mesh(Test_mesh):
    n_cells = 5
    mesh_type = "finite_difference"

    # Expected Paramaters
    expected_xcell_center = np.array([0, 0.25, 0.5, 0.75, 1])
    expected_n_cells = 5
    expected_laplacian = np.array(
        [
            [0, -0.5, 0, 0, 0],
            [0.5, 0, -0.5, 0, 0],
            [0, 0.5, 0, -0.5, 0],
            [0, 0, 0.5, 0, -0.5],
            [0, 0, 0, 0.5, 0],
        ]
    )
    expected_left_dirichlet_laplacian = np.array(
        [
            [0, 0, 0, 0, 0],
            [0.5, 0, -0.5, 0, 0],
            [0, 0.5, 0, -0.5, 0],
            [0, 0, 0.5, 0, -0.5],
            [0, 0, 0, 0.5, 0],
        ]
    )

    expected_right_laplacian = np.array(
        [
            [0, -0.5, 0, 0, 0],
            [0.5, 0, -0.5, 0, 0],
            [0, 0.5, 0, -0.5, 0],
            [0, 0, 0.5, 0, -0.5],
            [0, 0, 1.5, -1, 0.5],
        ]
    )
    expected_left_dirichlet_boundary_condition_array = np.array([0, 0, 0, 0, 0])

    expected_left_dirichlet_phi = np.array([5, 0, 0, 0, 0])

    @pytest.fixture
    def mesh_fixture(self):
        return mesher.linear_convection_mesh(
            self.x_range,
            n_cells=self.n_cells,
            mesh_type=self.mesh_type,
            convection_coefficient=1,
            discretization_type="central",
        )

    def test_central_laplacian(self, mesh_fixture):
        np.testing.assert_array_equal(mesh_fixture.laplacian, self.expected_laplacian)

    def test_central_left_dirichlet_laplacian(self, mesh_fixture):
        mesh_fixture.set_dirichlet_boundary("left", 5)
        np.testing.assert_array_equal(
            mesh_fixture.laplacian,
            self.expected_left_dirichlet_laplacian,
        )

    def test_central_left_dirichlet_boundary_condtion(self, mesh_fixture):
        mesh_fixture.phi.set_phi(0)
        mesh_fixture.set_dirichlet_boundary("left", 5)
        np.testing.assert_array_equal(
            mesh_fixture.boundary_condition_array,
            self.expected_left_dirichlet_boundary_condition_array,
        )

    def test_central_left_dirichlet_phi(self, mesh_fixture):
        mesh_fixture.phi.set_phi(0)
        mesh_fixture.set_dirichlet_boundary("left", 5)
        np.testing.assert_array_equal(
            mesh_fixture.phi.get_phi(),
            self.expected_left_dirichlet_phi,
        )

    @pytest.mark.skip(reason="unsure if discritization was performed correctly")
    def test_right_laplacian(self, mesh_fixture):
        mesh_fixture.set_right_boundary()
        np.testing.assert_array_equal(
            mesh_fixture.laplacian,
            self.expected_right_laplacian,
        )

    def test_discritization_error(self):
        with pytest.raises(ValueError):
            mesher.linear_convection_mesh(
                self.x_range,
                n_cells=self.n_cells,
                mesh_type=self.mesh_type,
                convection_coefficient=1,
                discretization_type="downwind",
            )


class Test_central_linear_convection_mesh_fv(Test_central_linear_convection_mesh):
    """Perform the same tests as teh centeral linear convection mesh for a finite volume"""

    n_cells = 4
    mesh_type = "finite_volume"

    # Expected Paramaters
    expected_xcell_center = np.array([0.125, 0.375, 0.625, 0.875])
    expected_n_cells = 4
    expected_laplacian = np.array(
        [
            [0, -0.5, 0, 0],
            [0.5, 0, -0.5, 0],
            [0, 0.5, 0, -0.5],
            [0, 0, 0.5, 0],
        ]
    )
    expected_left_dirichlet_laplacian = np.array(
        [
            [-0.5, -0.5, 0, 0],
            [0.5, 0, -0.5, 0],
            [0, 0.5, 0, -0.5],
            [0, 0, 0.5, 0],
        ]
    )

    expected_right_dirichlet_laplacian = np.array(
        [
            [0, -0.5, 0, 0],
            [0.5, 0, -0.5, 0],
            [0, 0.5, 0, -0.5],
            [0, 0, 0.5, 0.5],
        ]
    )
    expected_left_dirichlet_boundary_condition_array = np.array([5, 0, 0, 0])

    expected_left_dirichlet_phi = np.array([0, 0, 0, 0])

    expected_right_dirichlet_laplacian = np.array(
        [
            [0, -0.5, 0, 0],
            [0.5, 0, -0.5, 0],
            [0, 0.5, 0, -0.5],
            [0, 0, 0.5, 0.5],
        ]
    )

    expected_right_dirichlet_boundary_condition_array = np.array([0, 0, 0, -5])

    def test_central_right_dirichlet_laplacian(self, mesh_fixture):
        mesh_fixture.set_dirichlet_boundary("right", 5)
        np.testing.assert_array_equal(
            mesh_fixture.laplacian,
            self.expected_right_dirichlet_laplacian,
        )

    def test_central_right_dirichlet_boundary_condtion(self, mesh_fixture):
        mesh_fixture.phi.set_phi(0)
        mesh_fixture.set_dirichlet_boundary("right", 5)
        np.testing.assert_array_equal(
            mesh_fixture.boundary_condition_array,
            self.expected_right_dirichlet_boundary_condition_array,
        )


###################################################################


class Test_linear_convection_mesh(Test_mesh):
    mesh_type = "finite_difference"
    expected_xcell_center = np.array([0, 1 / 3, 2 / 3, 1])
    expected_delta_x: float = 1 / 3

    @pytest.fixture
    def mesh_fixture(self):
        return mesher.linear_convection_mesh(
            self.x_range,
            n_cells=self.n_cells,
            mesh_type=self.mesh_type,
            convection_coefficient=1,
        )

    expected_laplacian = np.array(
        [[-1, 0, 0, 0], [1, -1, 0, 0], [0, 1, -1, 0], [0, 0, 1, -1]]
    )

    expected_left_dirichlet_laplacian = np.array(
        [[0, 0, 0, 0], [1, -1, 0, 0], [0, 1, -1, 0], [0, 0, 1, -1]]
    )
    expected_left_boundary_condition_array = np.array([0, 0, 0, 0])

    def test_initiate_laplacian(self, mesh_fixture):
        # mesh_fixture.set_centeral_
        np.testing.assert_allclose(
            actual=mesh_fixture.laplacian,
            desired=self.expected_laplacian,
            atol=0.000001,
        )

    def test_initiate_differentation_matrix_neg(self):
        """Test that a negative convcection coefficent shows unsupported error"""
        with pytest.raises(ValueError):
            mesher.linear_convection_mesh(
                x=[0, 1], n_cells=4, convection_coefficient=-1
            )

    def test_set_phi(self, mesh_fixture):
        mesh_fixture.phi.set_phi(phi=[1, 2, 4, 5])
        np.testing.assert_equal(mesh_fixture.phi.get_phi(), [1, 2, 4, 5])

    def test_set_phi_float(self, mesh_fixture):
        mesh_fixture.phi.set_phi(phi=5.0)
        np.testing.assert_equal(mesh_fixture.phi.get_phi(), [5, 5, 5, 5])

    def test_set_phi_int(self, mesh_fixture):
        mesh_fixture.phi.set_phi(phi=5)
        np.testing.assert_equal(mesh_fixture.phi.get_phi(), [5, 5, 5, 5])

    def test_phi_nparray(self, mesh_fixture):
        expected = np.array([5, 5, 5, 5])
        mesh_fixture.phi.set_phi(expected.tolist())
        np.testing.assert_equal(mesh_fixture.phi.get_phi(), [5, 5, 5, 5])

    def test_set_phi_float_lenght(self, mesh_fixture):
        mesh_fixture.phi.set_phi(phi=5.0)
        with np.testing.assert_raises(AssertionError):
            np.testing.assert_array_equal(mesh_fixture.phi.get_phi(), [5, 5, 5, 5, 5])

    def test_set_phi_wrong_shape_error(self, mesh_fixture):
        """Ensure the shape of phi matches the shape of x"""
        with pytest.raises(ValueError):
            mesh_fixture.phi.set_phi([1, 2, 3, 4, 5])

    def test_set_phi_wrong_type(self, mesh_fixture):
        with pytest.raises(TypeError):
            mesh_fixture.phi.set_phi("string")

    def test_convection_coeff_saved(self, mesh_fixture):
        assert mesh_fixture.convection_coefficent == 1

    def test_left_dirichlet_boundary(self, mesh_fixture):
        mesh_fixture.set_dirichlet_boundary("left", 5)
        np.testing.assert_array_equal(
            mesh_fixture.boundary_condition_array,
            self.expected_left_boundary_condition_array,
        )

    def test_left_dirichlet_differentiaton_matrix(self, mesh_fixture):
        mesh_fixture.set_dirichlet_boundary("left", 5)
        np.testing.assert_array_equal(
            mesh_fixture.laplacian,
            self.expected_left_dirichlet_laplacian,
        )

    @pytest.mark.skip(reason="implementing right bc")
    def test_right_dirichlet_boundary(self, mesh_fixture):
        """because right bc is not implemented yet ensure a value error is raised"""
        with pytest.raises(ValueError):
            mesh_fixture.set_dirichlet_boundary("right", 5)


class Test_heat_diffusion_mesh(Test_mesh):
    @pytest.fixture
    def mesh_fixture(self):
        return heat_diffusion_mesh(
            self.x_range, n_cells=self.n_cells, mesh_type=self.mesh_type
        )

    expected_laplacian = np.array(
        [[-2, 1, 0, 0], [1, -2, 1, 0], [0, 1, -2, 1], [0, 0, 1, -2]]
    )

    expected_boundary_condition_left_dirichlet = np.array([100, 0, 0, 0])
    expected_boundary_condition_left_dirichlet_d2matrix = np.array(
        [[-3, 1, 0, 0], [1, -2, 1, 0], [0, 1, -2, 1], [0, 0, 1, -2]]
    )

    expected_boundary_condition_right_dirichlet = np.array([0, 0, 0, 100])
    expected_boundary_condition_right_dirichlet_d2matrix = np.array(
        [[-2, 1, 0, 0], [1, -2, 1, 0], [0, 1, -2, 1], [0, 0, 1, -3]]
    )

    # expected neumann with q = 50
    expected_boundary_condition_left_neumann = np.array([12.5, 0, 0, 0])
    expected_boundary_condition_left_neumann_d2matrix = np.array(
        [[-1, 1, 0, 0], [1, -2, 1, 0], [0, 1, -2, 1], [0, 0, 1, -2]]
    )

    expected_boundary_condition_right_neumann = np.array([0, 0, 0, 12.5])
    expected_boundary_condition_right_neumann_d2matrix = np.array(
        [[-2, 1, 0, 0], [1, -2, 1, 0], [0, 1, -2, 1], [0, 0, 1, -1]]
    )

    def test_set_diffusion_const(self, mesh_fixture):
        mesh_fixture.set_thermal_diffusivity(4)
        assert mesh_fixture.thermal_diffusivity == 4

    def test_initiate_laplacian(self, mesh_fixture):
        np.testing.assert_allclose(
            actual=mesh_fixture.laplacian,
            desired=self.expected_laplacian,
            atol=0.000001,
        )

    def test_set_internal_initial_temperature(self, mesh_fixture):
        mesh_fixture.set_cell_temperature(20)
        np.testing.assert_allclose(
            actual=mesh_fixture.temperature, desired=np.full(self.n_cells, 20)
        )

    def test_boundary_condition_initialized(self, mesh_fixture):
        np.testing.assert_allclose(
            actual=mesh_fixture.boundary_condition_array,
            desired=np.zeros(self.n_cells),
        )

    def test_left_dirclet_BC_array(self, mesh_fixture):
        mesh_fixture.set_dirichlet_boundary("left", 50)

        np.testing.assert_allclose(
            actual=mesh_fixture.boundary_condition_array,
            desired=self.expected_boundary_condition_left_dirichlet,
        )

    def test_left_dirclet_D2_Matrix(self, mesh_fixture):
        mesh_fixture.set_dirichlet_boundary("left", 50)

        np.testing.assert_allclose(
            actual=mesh_fixture.laplacian,
            desired=self.expected_boundary_condition_left_dirichlet_d2matrix,
        )

    def test_right_dirclet_BC_array(self, mesh_fixture):
        mesh_fixture.set_dirichlet_boundary("right", 50)
        np.testing.assert_allclose(
            actual=mesh_fixture.boundary_condition_array,
            desired=self.expected_boundary_condition_right_dirichlet,
        )

    def test_right_dirclet_D2_Matrix(self, mesh_fixture):
        mesh_fixture.set_dirichlet_boundary("right", 50)

        np.testing.assert_allclose(
            actual=mesh_fixture.laplacian,
            desired=self.expected_boundary_condition_right_dirichlet_d2matrix,
        )

    def test_left_neumann_BC_array(self, mesh_fixture):
        mesh_fixture.set_neumann_boundary("left", 50)

        np.testing.assert_allclose(
            actual=mesh_fixture.boundary_condition_array,
            desired=self.expected_boundary_condition_left_neumann,
        )

    def test_left_neumann_D2_Matrix(self, mesh_fixture):
        mesh_fixture.set_neumann_boundary("left", 50)

        np.testing.assert_allclose(
            actual=mesh_fixture.laplacian,
            desired=self.expected_boundary_condition_left_neumann_d2matrix,
        )

    def test_right_neumann_BC_array(self, mesh_fixture):
        mesh_fixture.set_neumann_boundary("right", 50)
        np.testing.assert_allclose(
            actual=mesh_fixture.boundary_condition_array,
            desired=self.expected_boundary_condition_right_neumann,
        )

    def test_dirchlet_BC_array_can_be_corrected(self, mesh_fixture):
        mesh_fixture.set_neumann_boundary("right", 50)
        mesh_fixture.set_dirichlet_boundary("right", 50)
        np.testing.assert_allclose(
            actual=mesh_fixture.boundary_condition_array,
            desired=self.expected_boundary_condition_right_dirichlet,
        )

    def test_right_neumann_D2_Matrix(self, mesh_fixture):
        mesh_fixture.set_neumann_boundary("right", 50)

        np.testing.assert_allclose(
            actual=mesh_fixture.laplacian,
            desired=self.expected_boundary_condition_right_neumann_d2matrix,
        )

    def test_unsuported_boundary_conndtion_raises(self, mesh_fixture):
        with pytest.raises(ValueError):
            mesh_fixture.set_neumann_boundary("lleft", 50)


# Test that a different x range can work
class Test_x_range(Test_heat_diffusion_mesh):
    x_range = [0, 4]
    expected_xcell_center = np.array([0.5, 1.5, 2.5, 3.5])
    expected_delta_x = 1
    expected_n_cells = 4
    expected_boundary_condition_left_neumann = np.array([50, 0, 0, 0])
    expected_boundary_condition_right_neumann = np.array([0, 0, 0, 50])


class Test_finite_difference(Test_heat_diffusion_mesh):
    n_cells = 5
    mesh_type = "finite_difference"

    # Expected Paramaters
    expected_xcell_center = np.array([0, 0.25, 0.5, 0.75, 1])
    expected_n_cells = 5
    expected_laplacian = np.array(
        [
            [-2, 1, 0, 0, 0],
            [1, -2, 1, 0, 0],
            [0, 1, -2, 1, 0],
            [0, 0, 1, -2, 1],
            [0, 0, 0, 1, -2],
        ]
    )

    expected_boundary_condition_left_dirichlet = np.array([0, 0, 0, 0, 0])
    expected_boundary_condition_left_dirichlet_d2matrix = np.array(
        [
            [0, 0, 0, 0, 0],
            [1, -2, 1, 0, 0],
            [0, 1, -2, 1, 0],
            [0, 0, 1, -2, 1],
            [0, 0, 0, 1, -2],
        ]
    )

    expected_boundary_condition_right_dirichlet = np.array([0, 0, 0, 0, 0])
    expected_boundary_condition_right_dirichlet_d2matrix = np.array(
        [
            [-2, 1, 0, 0, 0],
            [1, -2, 1, 0, 0],
            [0, 1, -2, 1, 0],
            [0, 0, 1, -2, 1],
            [0, 0, 0, 0, 0],
        ]
    )

    # neuiman should be 2*q*delta_x when a point is on the boudnary
    expected_boundary_condition_left_neumann = np.array([25, 0, 0, 0, 0])
    expected_boundary_condition_left_neumann_d2matrix = np.array(
        [
            [-2, 2, 0, 0, 0],
            [1, -2, 1, 0, 0],
            [0, 1, -2, 1, 0],
            [0, 0, 1, -2, 1],
            [0, 0, 0, 1, -2],
        ]
    )

    expected_boundary_condition_right_neumann = np.array([0, 0, 0, 0, 25])

    expected_boundary_condition_right_neumann_d2matrix = np.array(
        [
            [-2, 1, 0, 0, 0],
            [1, -2, 1, 0, 0],
            [0, 1, -2, 1, 0],
            [0, 0, 1, -2, 1],
            [0, 0, 0, 2, -2],
        ]
    )


@pytest.fixture
def five_cell_mesh():
    return heat_diffusion_mesh(x=[0, 1], n_cells=5, mesh_type="finite_difference")


def test_finite_difference_dirichlet_set_temperature(five_cell_mesh):
    """
    Regression test for issue 10
    Description
    When using the finite difference, setting the dirichlet boundary
    does not enforce the temeprature
    """

    five_cell_mesh.set_cell_temperature(20)
    five_cell_mesh.set_dirichlet_boundary("left", 50)
    five_cell_mesh.set_dirichlet_boundary("right", 45)
    expected_temperature = np.array([50, 20, 20, 20, 45])
    np.testing.assert_allclose(
        actual=five_cell_mesh.temperature, desired=expected_temperature
    )


@pytest.mark.xfail(reason="feature improvment recomendation")
def test_finite_difference_dirichlet_overwrite_temperature(five_cell_mesh):
    """
    Regression test for issue 10
    Description
    Initial resolution for issue 10 allowed for setting the boundary temperatures after the initial temperatures
    However if the internal temperatures are set after the boundarys will be overwritten
    Below is a testcase of optimal behavior
    """

    five_cell_mesh.set_dirichlet_boundary("left", 50)
    five_cell_mesh.set_cell_temperature(20)
    five_cell_mesh.set_dirichlet_boundary("right", 45)

    expected_temperature = np.array([50, 20, 20, 20, 45])
    np.testing.assert_allclose(
        actual=five_cell_mesh.temperature, desired=expected_temperature
    )


def test_init():
    from solver import mesher

    with patch.object(mesher, "main", MagicMock()) as mock_main:
        with patch.object(mesher, "__name__", "__main__"):
            mesher.init()
            mock_main.assert_called_once()


class Test_upwind_linear_convection_mesh_finite_volume(Test_linear_convection_mesh):
    n_cells = 4
    mesh_type = "finite_volume"
    discritization_type = "upwind"
    expected_delta_x = 0.25
    # Expected Paramaters
    expected_xcell_center = np.array([0.125, 0.375, 0.625, 0.875])
    expected_n_cells = 4
    expected_laplacian = np.array(
        [
            [-1, 0, 0, 0],
            [1, -1, 0, 0],
            [0, 1, -1, 0],
            [0, 0, 1, -1],
        ]
    )
    expected_left_dirichlet_laplacian = np.array(
        [
            [-1, 0, 0, 0],
            [1, -1, 0, 0],
            [0, 1, -1, 0],
            [0, 0, 1, -1],
        ]
    )

    expected_right_laplacian = np.array(
        [
            [-1, 0, 0, 0],
            [1, -1, 0, 0],
            [0, 1, -1, 0],
            [0, 0, 1, -1],
        ]
    )

    expected_left_boundary_condition_array = np.array([5, 0, 0, 0])

    expected_left_dirichlet_phi = np.array([0, 0, 0, 0])


class Test_cell_phi:
    @pytest.fixture
    def one_d_phi_finite_dif(self):
        return mesher.cell_phi(n_cells=3, dim=1, mesh_type="finite_difference")

    def test_1d_phi(self):
        expected = np.array([0, 0, 0, 0])
        actual = mesher.cell_phi(n_cells=4, dim=1, mesh_type="finite_volume").get_phi()
        np.testing.assert_array_equal(x=actual, y=expected)

    def test_1d_phi_listcell(self):
        expected = np.array([0, 0, 0, 0])
        actual = mesher.cell_phi(
            n_cells=[4], dim=1, mesh_type="finite_volume"
        ).get_phi()
        np.testing.assert_array_equal(x=actual, y=expected)

    def test_1d_phi_left_dirichlet(self, one_d_phi_finite_dif):
        expected = np.array([10, 0, 0])
        one_d_phi_finite_dif.set_dirichlet_boundary("left", 10)
        actual = one_d_phi_finite_dif.get_phi()
        np.testing.assert_array_equal(x=actual, y=expected)

    def test_2d_phi(self):
        expected = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0]])
        actual = mesher.cell_phi(
            n_cells=[3, 4], dim=2, mesh_type="finite_volume"
        ).get_phi()
        np.testing.assert_array_equal(x=actual, y=expected)

    @pytest.mark.parametrize("mesh", ["finite_volume", "finite_difference"])
    def test_2d_set_phi(self, mesh):
        expected = np.array([[10, 10, 10], [10, 10, 10], [10, 10, 10], [10, 10, 10]])
        actual = mesher.cell_phi(n_cells=[3, 4], dim=2, mesh_type=mesh)
        actual.set_phi(10)
        np.testing.assert_array_equal(x=actual.get_phi(), y=expected)

    def test_2d_phi_left_dirichlet(self):
        expected = np.array([[10, 0, 0], [10, 0, 0], [10, 0, 0], [10, 0, 0]])
        actual = mesher.cell_phi(n_cells=[3, 4], dim=2, mesh_type="finite_difference")
        actual.set_dirichlet_boundary(side="left", phi=10)

        np.testing.assert_array_equal(x=actual.get_phi(), y=expected)

    def test_2d_phi_right_dirichlet(self):
        expected = np.array([[0, 0, 10], [0, 0, 10], [0, 0, 10], [0, 0, 10]])
        actual = mesher.cell_phi(n_cells=[3, 4], dim=2, mesh_type="finite_difference")
        actual.set_dirichlet_boundary(side="right", phi=10)

        np.testing.assert_array_equal(x=actual.get_phi(), y=expected)

    def test_2d_phi_top_dirichlet(self):
        expected = np.array([[10, 10, 10], [0, 0, 0], [0, 0, 0], [0, 0, 0]])
        actual = mesher.cell_phi(n_cells=[3, 4], dim=2, mesh_type="finite_difference")
        actual.set_dirichlet_boundary(side="top", phi=10)

        np.testing.assert_array_equal(x=actual.get_phi(), y=expected)

    def test_2d_phi_bottom_dirichlet(self):
        expected = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 0], [10, 10, 10]])
        actual = mesher.cell_phi(n_cells=[3, 4], dim=2, mesh_type="finite_difference")
        actual.set_dirichlet_boundary(side="bottom", phi=10)

        np.testing.assert_array_equal(x=actual.get_phi(), y=expected)

    def test_phi_mesh_type_validate(self):
        with pytest.raises(ValueError):
            mesher.cell_phi(n_cells=1, dim=1, mesh_type="unsuported")


class Test_boundary_condition_array:
    @pytest.mark.parametrize(
        "side,expected",
        [
            ("left", np.array([60, 0, 0])),
            ("top", np.array([60, 0, 0])),
            ("right", np.array([0, 0, 60])),
            ("bottom", np.array([0, 0, 60])),
        ],
    )
    def test_boundary_condtion_dirichlet(self, side, expected):
        actual = mesher.boundary_condition(n_cells=3, mesh_type="finite_volume")
        actual.set_dirichlet_boundary(side, 30)
        np.testing.assert_array_equal(x=actual.get_array(), y=expected)


class Test_differentiation_matrix:
    left_dirichlet = np.array([[-3, 1, 0], [1, -2, 1], [0, 1, -2]])
    right_dirichlet = np.array([[-2, 1, 0], [1, -2, 1], [0, 1, -3]])

    @pytest.mark.parametrize(
        "side,expected",
        [
            ("left", left_dirichlet),
            ("top", left_dirichlet),
            ("right", right_dirichlet),
            ("bottom", right_dirichlet),
        ],
    )
    def test_differentiation_matrix_dirichlet(self, side, expected):
        actual = mesher.differentiation_matrix(n_cells=3)
        actual.set_dirichlet_boundary(side, "finite_volume")
        np.testing.assert_array_equal(x=actual.get_matrix(), y=expected)

    left_neumann = np.array([[-1, 1, 0], [1, -2, 1], [0, 1, -2]])
    right_neumann = np.array([[-2, 1, 0], [1, -2, 1], [0, 1, -1]])

    @pytest.mark.parametrize(
        "side,expected",
        [
            ("left", left_neumann),
            ("top", left_neumann),
            ("right", right_neumann),
            ("bottom", right_neumann),
        ],
    )
    def test_boundary_condtion_neumann(self, side, expected):
        actual = mesher.differentiation_matrix(n_cells=3)
        actual.set_neumann_boundary(side, "finite_volume")
        np.testing.assert_array_equal(x=actual.get_matrix(), y=expected)



import numpy as np


class SteadySolver(object):
    def solve(self, laplacian, boundary_condition_array):
        return np.linalg.solve(laplacian, -boundary_condition_array)


class CartesianSolver:
    def __init__(self, solver=SteadySolver()):
        self.solver = solver

    #     """A solver to solve a cartesian mesh"""
    def solve(self, **kwags):
        return self.solver.solve(**kwags)



import numpy as np
import logging
from solver.mesher import side_selector
from matplotlib import pyplot as plt

# create logging configuration
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# Create a console handler
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.INFO)

# Set the formatter for the console handler
formatter = logging.Formatter(
    "%(name)s:%(levelname)s:%(funcName)s:%(message)s",
)
console_handler.setFormatter(formatter)

# Add the console handler to the logger
logger.addHandler(console_handler)


class MeshReshaper:
    """Object to store a mesh shape and return a long or wide mesh"""

    def __init__(self, x_cells, y_cells):
        """Store the mesh dimensions"""
        self.x_cells = x_cells
        self.y_cells = y_cells

    def to_long(self, array):
        """Return a long array"""
        return np.reshape(array, self.x_cells * self.y_cells)

    def to_wide(self, array):
        """Return a wide array"""
        return np.reshape(array, (self.y_cells, self.x_cells))


class EnergyBalance:
    """Object to calculate the energy balance for a mesh"""

    def __init__(self, mesh, side_selector=side_selector()):
        self.mesh = mesh
        self.side_selector = side_selector
        self.x_cells = self.mesh.n_cells[0]
        self.y_cells = self.mesh.n_cells[1]
        self.x_width = self.mesh.grid["x_grid"].cell_width
        self.y_width = self.mesh.grid["y_grid"].cell_width
        self.phi = self.mesh.phi.get_phi()
        self.generation = self.set_generation()
        self.x_flux = self.set_x_flux()
        self.y_flux = self.set_y_flux()
        self.cell_flux = self.set_cell_flux()

    def set_generation(self):
        generation = self.mesh.generation.reshape(self.phi.shape) * (
            self.y_width * self.x_width
        )
        return generation

    def set_x_flux(self):
        x_flux_diff_matrix = np.zeros((self.x_cells, self.x_cells))
        y_identity = np.identity(self.y_cells)

        for side in ["left", "right"]:
            boundary_index = self.side_selector.boundary_index(side)
            differentiation_value = self.differntiation_value(side)
            x_flux_diff_matrix[boundary_index, boundary_index] = differentiation_value

        d2_x = np.kron(y_identity, x_flux_diff_matrix)

        self.bray_x_flux = (
            ((d2_x @ self.phi.flatten()) + self.mesh.x_bc_reshape.flatten())
            * (self.mesh.conductivity * self.y_width / self.x_width)
        ).reshape(self.phi.shape)
        return (
            ((d2_x @ self.phi.flatten()) + self.mesh.x_bc_reshape.flatten())
            * (self.mesh.conductivity * self.y_width / self.x_width)
        ).reshape(self.phi.shape)

    def set_y_flux(self):
        y_flux_diff_matrix = np.zeros((self.y_cells, self.y_cells))
        x_identity = np.identity(self.x_cells)

        for side in ["top", "bottom"]:
            boundary_index = self.side_selector.boundary_index(side)
            differentiation_value = self.differntiation_value(side)
            y_flux_diff_matrix[boundary_index, boundary_index] = differentiation_value

        d2_y = np.kron(y_flux_diff_matrix, x_identity)

        return (
            ((d2_y @ self.phi.flatten()) + self.mesh.y_bc_reshape.flatten())
            * (self.mesh.conductivity * self.x_width / self.y_width)
        ).reshape(self.phi.shape)

    def set_cell_flux(self):
        """Determine the heat flux for each cell."""
        x_identity = np.identity(self.x_cells)
        y_identity = np.identity(self.y_cells)

        d2y = np.kron(self.mesh.d2y_unscaled, x_identity)

        cell_y_flux = (
            d2y @ self.phi.flatten() + self.mesh.y_bc_reshape.flatten()
        ).reshape(self.phi.shape) * (
            self.mesh.conductivity * self.x_width / self.y_width
        )

        d2x = np.kron(y_identity, self.mesh.d2x_unscaled)
        cell_x_flux = (
            d2x @ self.phi.flatten() + self.mesh.x_bc_reshape.flatten()
        ).reshape(self.phi.shape) * (
            self.mesh.conductivity * self.y_width / self.x_width
        )

        return cell_x_flux + cell_y_flux + self.generation

    # cell_flux_se = np.sum((cell_flux**2))

    def flux(self, side: str):
        if side == "left":
            return np.sum(self.x_flux[:, 0])
        if side == "right":
            return np.sum(self.x_flux[:, -1])
        if side == "top":
            return np.sum(self.y_flux[0, :])
        if side == "bottom":
            return np.sum(self.y_flux[-1, :])
        if side == "cells":
            return np.sum((self.cell_flux**2))
        if side == "generation":
            return np.sum(self.generation)
        if side == "all":
            total_flux = np.sum(self.y_flux + self.x_flux + self.generation)
            logger.info(
                f"\n Left Flux: {np.sum(self.x_flux[:,0])} W  \
                \n Right Flux: {np.sum(self.x_flux[:,-1])} W  \
                \n Bottom Flux: {np.sum(self.y_flux[-1,:])} W  \
                \n Generation: {np.sum(self.generation)} W \
                \n Top Flux: {np.sum(self.y_flux[0,:])} W  \
                \n Cell Sum Squared Error: {np.sum((self.cell_flux**2))} W \
                \n Flux leaving boundarys: {np.sum(self.y_flux+ self.x_flux)}W \
                \n Total Flux: {total_flux}"
            )

            return total_flux

    def differntiation_value(self, side: str):
        if self.mesh.boundary_condition_dict[side] == "dirichlet":
            return -2
        elif self.mesh.boundary_condition_dict[side] == "neumann":
            return 0
        else:
            recieved = self.mesh.boundary_condition_dict[side]
            raise ValueError(f"Boundary conditions needed, got {recieved}")


class Parser:
    def parse(self, string: str):
        x = string.split("_")
        return x[0]


class Plotter:
    """A class to plot a 2d Cartesian mesh at various timepoints"""

    def __init__(self, mesh):
        self.mesh = mesh

    def transient_plotter(self, data_list, name, phi_min=10, phi_max=30):
        """plot 4 seperate transient temperature profiles

        args
        data_list: A list of dictionarys that have the time point as the key and phi values as the values
        mesh: the mesh
        """
        fig, axes = plt.subplots(4, figsize=(10, 12))
        fig2, axes2 = plt.subplots(4, figsize=(10, 12))
        list_length = len(data_list)
        for row, ax, ax2 in zip(
            [0, round(list_length / 4), round(3 * list_length / 4), list_length - 1],
            axes.flat,
            axes2.flat,
        ):
            # for time, data in data_list[row].items():
            time = data_list[row]["time"]
            data = data_list[row]["phi"]
            x_cords = self.mesh.grid["x_grid"].cell_cordinates
            y_cords = self.mesh.grid["y_grid"].cell_cordinates
            logger.debug(row)
            xv, yv = np.meshgrid(x_cords, y_cords)

            tmin = phi_min
            tmax = phi_max
            y_cells = data.shape[0]

            # Plot the temperature profiles for each time step
            im = ax.pcolormesh(xv, yv, data, vmax=tmax, vmin=tmin)
            ax.set_title(f"Temperature Profile at Time {time} s")

            # Plot the the temperature distributions for various horizontal lines
            group = [1, 2, 3]
            midline = round(y_cells / 2)
            top_quarter_line = round(y_cells / 4)
            bottom_quarter_line = round(3 * y_cells / 4)

            line_list = [midline, top_quarter_line, bottom_quarter_line]
            color_group = {1: "red", 2: "blue", 3: "green"}
            color_group_name = {1: "midline", 2: "top", 3: "bottom"}

            for g in group:
                line_row = line_list[g - 1]
                ax.axhline(y=(1 - (line_row / y_cells)), color=color_group[g])
                ax2.plot(
                    x_cords,
                    data[line_row, :],
                    color=color_group[g],
                    label=color_group_name[g],
                )

                ax2.axis((0, 5, tmin, tmax + 1))
                ax2.set_ylabel("temperature (celcius)")
                ax2.set_title(f"midpoint temperature at time {time} s")
                ax2.legend()

                fig.subplots_adjust(bottom=0.1, top=0.9, hspace=0.4)
                fig.colorbar(
                    im, ax=axes.ravel().tolist(), label="temperature (celcius)"
                )

                fig2.subplots_adjust(bottom=0.1, top=0.9, hspace=0.5)
                fig.savefig(f"{name}_temperature_profile.png")
                fig2.savefig(f"{name}_line_distributions.png")


class VelocityRounder:
    def t_final(self, velocity):
        t_final = round(10 / velocity)
        return t_final

    def time_step_size(self, velocity):
        return self.rounder(velocity, 0.01)

    def rounder(self, velocity, scale):
        if round(velocity / scale) == 0:
            return round(0.01 / velocity)
        else:
            return 1 / round(velocity / scale)



from solver import utilities
from solver import cartesian_mesh
import numpy as np
import pytest
import math
import logging
from unittest.mock import PropertyMock


class TestMeshReshaper:
    @pytest.fixture
    def long_array(self):
        return np.arange(10)

    @pytest.fixture
    def wide_array(self, long_array):
        return np.reshape(long_array, (2, 5))

    def test_reshape_long(self, long_array, wide_array):
        actual = utilities.MeshReshaper(x_cells=5, y_cells=2).to_long(wide_array)
        expected = long_array

        np.testing.assert_array_equal(x=actual, y=expected)

    def test_reshape_wide(self, long_array, wide_array):
        actual = utilities.MeshReshaper(x_cells=5, y_cells=2).to_wide(long_array)
        expected = wide_array

        np.testing.assert_array_equal(x=actual, y=expected)


class TestEnergyBalance:
    @pytest.fixture
    def mock_mesh(self, mocker):
        """
        Create to create a mock mesh object for use in testing the solver.
        Mesh configuration
        N_elements =
        Left boundary = Dirchilet at 50c
        Right boundary = neuiman with q=0
        :return: mesh
        """

        mesh = mocker.MagicMock()
        mesh.phi.phi = np.array(
            [
                [28.72076498, 27.94615599, 28.72076498],
                [29.70707761, 29.46041556, 29.70707761],
                [29.93022914, 29.86469587, 29.93022914],
                [29.98686165, 29.97407644, 29.98686165],
            ]
        )
        mesh.x_bc_reshape = np.array(
            [[60.0, 0.0, 60.0], [60.0, 0.0, 60.0], [60.0, 0.0, 60.0], [60.0, 0.0, 60.0]]
        )

        mesh.y_bc_reshape = np.array(
            [[-5.0, -5.0, -5.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [60.0, 60.0, 60.0]]
        )
        mesh.phi.get_phi = mocker.MagicMock(return_value=mesh.phi.phi)

        mesh.boundary_condition_dict = {
            "left": "dirichlet",
            "right": "dirichlet",
            "bottom": "dirichlet",
            "top": "neuimann",
        }

        mesh.grid["x_grid"].cell_width = 5
        # mocker.MagicMock(return_value = [4,5])
        # mesh.grid["y_grid"].cell_width = 3

        mesh.n_cells = [3, 4]
        mesh.conductivity = 1
        return mesh

    def test_my_mock_mesh(self, mock_mesh):
        expected = np.array(
            [
                [28.72076498, 27.94615599, 28.72076498],
                [29.70707761, 29.46041556, 29.70707761],
                [29.93022914, 29.86469587, 29.93022914],
                [29.98686165, 29.97407644, 29.98686165],
            ]
        )
        actual = mock_mesh.phi.get_phi()
        np.testing.assert_array_equal(x=actual, y=expected)

    @pytest.fixture
    def cartesian_mesh(self):
        mesh = cartesian_mesh.CartesianMesh(
            dimensions=2, n_cells=[3, 4], cordinates=[(0, 1), (0, 2)], conductivity=1
        )

        mesh.set_dirichlet_boundary(side="left", phi=30)
        mesh.set_dirichlet_boundary(side="right", phi=30)
        mesh.set_dirichlet_boundary(side="bottom", phi=30)
        mesh.set_neumann_boundary(side="top", flux=-10)
        mesh.phi.set_phi(
            [
                [28.72076498, 27.94615599, 28.72076498],
                [29.70707761, 29.46041556, 29.70707761],
                [29.93022914, 29.86469587, 29.93022914],
                [29.98686165, 29.97407644, 29.98686165],
            ]
        )
        return mesh

    @pytest.mark.parametrize(
        "side,expected_fixture",
        [
            ("left", 4.965199833748016),
            ("right", 4.965199833748016),
            ("bottom", 0.06960033250395742),
            ("top", -10),
            ("all", 0.00000001),
            ("cells", 0.00000001),
        ],
    )
    def test_left_energy_balance(self, cartesian_mesh, side, expected_fixture):
        actual = utilities.EnergyBalance(mesh=cartesian_mesh).flux(side)
        expected = expected_fixture
        logging.debug(f"actual:{actual}, expected:{expected}")
        assert math.isclose(actual, expected, abs_tol=0.000001)


class TestAxisParse:
    @pytest.mark.parametrize("input_str,expected", [("x_grid", "x"), ("y_grid", "y")])
    def test_axis_parse(self, input_str, expected):
        actual = utilities.Parser().parse(input_str)
        assert actual == expected


class TestVelocityRounder:
    @pytest.mark.parametrize(
        "velocity,t_final",
        [
            (0.000001, 10000000),
            (0.00001, 1000000),
            (0.01, 1000),
            (0.1, 100),
            (1, 10),
            (10, 1),
        ],
    )
    def test_max_velocity(self, velocity, t_final):
        actual = utilities.VelocityRounder().t_final(velocity)
        assert actual == t_final

    @pytest.mark.parametrize(
        "velocity,expected",
        [
            (0.000001, 10000),
            (0.00001, 1000),
            (0.01, 1),
            (0.1, 0.1),
            (1, 0.01),
            (10, 0.001),
        ],
    )
    def test_time_step_size(self, velocity, expected):
        actual = utilities.VelocityRounder().time_step_size(velocity)
        assert actual == expected



from solver.mesher import (
    grid,
    differentiation_matrix,
    boundary_condition,
    side_selector,
    cell_phi,
)

from solver.utilities import Parser
from typing import Sequence, Tuple, List
import numpy as np
import logging

# create logging configuration
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# Create a console handler
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.INFO)

# Set the formatter for the console handler
formatter = logging.Formatter(
    "%(name)s:%(levelname)s:%(funcName)s:%(message)s",
)
console_handler.setFormatter(formatter)

# Add the console handler to the logger
logger.addHandler(console_handler)

# from cartesian_solver import CartesianSolver


class CartesianMesh:
    """
    A cartesian mesh up to 2D.

    Atributes:
    dimensions: int =  Number of dimensions (1d or 2d)
    n_cells:List[int]: Number of cells to discritize each dimension
    cordinates:Sequence[Tuple[float, float]] = The bounds each dimension
    mesh_type: str = The mesh type. currently only finite_volume.

    Example:
    To create a 2d mesh with
        3 cells in the x axis going from 0 to 1
        4 cells in the y axis going from 0 to 2
        cartesian_mesh(
            dimensions = 2
            n_cells = [3, 4]
            cordinates = [(0,1), (0,2)]
    """

    def __init__(
        self,
        dimensions: int = 2,
        n_cells: List[int] = [4, 4],
        cordinates: Sequence[Tuple[float, float]] = [(0, 1), (0, 1)],
        mesh_type: str = "finite_volume",
        conductivity: float = 1,
        diffusivity: float = 1,
    ) -> None:
        """
        Init the cartesian mesh object.

        Args:
            dimensions:int = Mesh dimensionality (default 2d)
            cordinates (list of tupples):list of each dimensions cordinates
            mesh_type: str = finite_volume (default) or finite_difference

        """
        self.validate_inputs(n_cells, cordinates, dimensions, mesh_type)
        self.n_cells = n_cells
        self.dimensions = dimensions
        self.cordinates = cordinates
        self.mesh_type = mesh_type
        self.diffusivity = diffusivity
        self.grid = self.initalize_grid()
        self.differentiation_matrix = self.initalize_differentiation_matrix()
        self.initalize_phi()
        self.boundary_condition = self.initalize_boundary_condition()
        self.set_laplacian()
        self.boundary_condition_dict: dict[str, str] = {}
        self.conductivity = conductivity

    def validate_inputs(
        self,
        n_cells: List[int] = [4, 4],
        cordinates: Sequence[Tuple[float, float]] = [(0, 1), (0, 1)],
        dimensions: int = 2,
        mesh_type: str = "finite_volume",
    ) -> None:
        """
        Validate the cartesian mesh inputs.

        Args:
            dimensions:int = Mesh dimensionality (default 2d)
            cordinates (list of tupples):list of each dimensions cordinates
            mesh_type: str = finite_volume (default) or finite_difference

        """
        # Implemented dimensions and mesh types
        self.implemented_dimensions: Tuple[str, str] = ("x", "y")
        self.implemented_mesh_types: str = "finite_volume"

        if dimensions > len(self.implemented_dimensions):
            raise ValueError("mesh dimesnionality not implemented")

        # Validate cordinates were given for each dimension
        if len(cordinates) != dimensions:
            raise ValueError("number of cordinates needs to match dimesnions")
        if len(n_cells) != dimensions:
            raise ValueError("legth of n_cells list needs to match dimension")

        if mesh_type not in self.implemented_mesh_types:
            raise ValueError(f"{mesh_type}: is not an implemented mesh")

    def initalize_grid(self):
        grid_dict = {}
        for index in range(0, self.dimensions):
            grid_dict[f"{self.implemented_dimensions[index]}_grid"] = grid(
                n_cells=self.n_cells[index],
                cordinates=self.cordinates[index],
                mesh_type=self.mesh_type,
            )
            # flip the y cordinates so the origin is in the top left corner
            if index == 1:
                grid_dict["y_grid"].cell_cordinates = np.flip(
                    grid_dict["y_grid"].cell_cordinates
                )
        return grid_dict

    def initalize_differentiation_matrix(self):
        """
        Create a differentiation matrix for each dimension.

        returns: a dictionary of differentiation matrix
        """
        differentiation_matrix_dict = {}
        for index in range(0, self.dimensions):
            differentiation_matrix_dict[
                f"{self.implemented_dimensions[index]}_differentiation_matrix"
            ] = differentiation_matrix(
                n_cells=self.n_cells[index],
            )

        return differentiation_matrix_dict

    def initalize_boundary_condition(self):
        """
        Create a boundary condition array for each dimension.

        returns: a dictionary of boundary conditions
        """
        boundary_condition_dict = {}
        for index in range(0, self.dimensions):
            boundary_condition_dict[
                f"{self.implemented_dimensions[index]}_boundary_condition_array"
            ] = boundary_condition(
                n_cells=self.n_cells[index], mesh_type=self.mesh_type
            )

        return boundary_condition_dict

    def initalize_phi(self):
        self.phi = cell_phi(self.n_cells, self.dimensions, self.mesh_type)

    def set_dirichlet_boundary(self, side: str, phi: float):
        """
        Set the dirichlet boundary.

        Updates differentiation matrix, and boundary condition array
        args:
        side: the side to set (left, right (1d) top, bottom (2d))
        phi: the value to set the boundary
        """
        axis = side_selector().axis(side)

        self.differentiation_matrix[
            f"{axis}_differentiation_matrix"
        ].set_dirichlet_boundary(side=side, mesh_type=self.mesh_type)

        self.boundary_condition[
            f"{axis}_boundary_condition_array"
        ].set_dirichlet_boundary(side, phi)

        self.boundary_condition_dict[side] = "dirichlet"
        self.set_laplacian()
        self.set_boundary_condition_array()
        self.phi.set_dirichlet_boundary(side, phi)
        self.generation = np.zeros(self.n_cells).flatten()

    def set_neumann_boundary(self, side: str, flux: float):
        """
        Set a neuman boundary.

        args:
        side: str = left, right (1d), top, bottom (2d)
        flux: float = flux into the boundary (negative if out)
        """
        axis = side_selector().axis(side)

        self.differentiation_matrix[
            f"{axis}_differentiation_matrix"
        ].set_neumann_boundary(side, self.mesh_type)

        self.boundary_condition[
            f"{axis}_boundary_condition_array"
        ].set_neumann_boundary(
            side=side, flux=flux, cell_width=self.grid[f"{axis}_grid"].cell_width
        )
        self.boundary_condition_dict[side] = "neumann"

        self.set_laplacian()
        self.set_boundary_condition_array()

    def set_laplacian(self):
        """Combine the differentiation matricies into a single matrix."""
        if self.dimensions == 1:
            # d2x = self.x_differentiation_matrix.get_matrix()
            self.laplacian = self.differentiation_matrix[
                "x_differentiation_matrix"
            ].get_matrix() * (self.diffusivity / self.grid["x_grid"].cell_width ** 2)
        elif self.dimensions == 2:
            self.d2x_unscaled = self.differentiation_matrix[
                "x_differentiation_matrix"
            ].get_matrix()
            self.d2y_unscaled = self.differentiation_matrix[
                "y_differentiation_matrix"
            ].get_matrix()
            d2x = self.d2x_unscaled * (
                self.diffusivity / self.grid["x_grid"].cell_width ** 2
            )

            d2y = self.d2y_unscaled * (
                self.diffusivity / self.grid["y_grid"].cell_width ** 2
            )
            Ix = np.identity(self.grid["x_grid"].n_cells)
            Iy = np.identity(self.grid["y_grid"].n_cells)
            self.laplacian = np.kron(Iy, d2x) + np.kron(d2y, Ix)

    def set_boundary_condition_array(self):
        """Combine boundary conditions into a single array."""
        if not hasattr(self, "generation"):
            self.generation = np.zeros(self.n_cells).flatten()
        if self.dimensions == 1:
            self.boundary_condition_array = self.boundary_condition[
                "x_boundary_condition_array"
            ].get_array() * (self.diffusivity / self.grid["x_grid"].cell_width ** 2)

        elif self.dimensions == 2:
            x_bc_array = self.boundary_condition[
                "x_boundary_condition_array"
            ].get_array()

            y_bc_array = self.boundary_condition[
                "y_boundary_condition_array"
            ].get_array()

            x_cells = self.grid["x_grid"].n_cells
            y_cells = self.grid["y_grid"].n_cells
            dx = self.grid["x_grid"].cell_width
            dy = self.grid["y_grid"].cell_width

            self.x_bc_reshape = x_bc_array.reshape(1, x_cells).repeat(y_cells, axis=0)
            self.y_bc_reshape = y_bc_array.reshape(y_cells, 1).repeat(x_cells, axis=1)
            logger.debug(f"generation in set bc {self.generation.shape}")
            square_boundary_condition = (
                (self.x_bc_reshape * (1 / dx**2))
                + (self.y_bc_reshape * (1 / dy**2))
            ) * self.diffusivity
            self.boundary_condition_array = square_boundary_condition.reshape(
                x_cells * y_cells
            ) + self.generation * (self.diffusivity / self.conductivity)

            logger.debug(
                f"combined boundary condition array{self.boundary_condition_array.shape}"
            )

    def set_generation(self, function):
        """
        set a generation function
        args:
        function: the generation function
            set_generation(lambda x : x**2) for 1d,
            set_generation(lambda x,y: 2x + y**2) for 2d

            note: the function must contain a variable for all active axies
            variables even if it is a zero i.e. (lambda x,y: 1x + 0*y)
            it can also be a named function:
            def my_function (x, y):
                return(2*x+3*y +5)
            set_generation(my_function)
        Behavior
            This function generates a dictionary for each grid axis in grid.items()
            using the slice function to evalueate the function in the grid space required
            (analagous to linspace). The function is then evaluated on the grid and flattened
            The set_boundary_condition_array is then called to ensure it is added to the boundary condition array
        Reference (https://stackoverflow.com/questions/22774726/numpy-evaluate-function-on-a-grid-of-points)
        """
        grid_dict = {}
        logger.debug(f"grid:{self.grid}")
        # Generate a dictionary for each axis as either a column or a row.
        # Analagous to meshgrid
        for i, (names, cordinate_value) in enumerate(self.grid.items()):
            grid_dict[(Parser().parse(names))] = cordinate_value.cell_cordinates[
                (None,) * i
                + (slice(None),)
                + (None,) * (len(self.grid.items()) - i - 1)
            ]

        logger.debug(f"grid_dict:{grid_dict}")
        self.generation = (function(**grid_dict)).flatten(order="F")
        logger.debug(f"generation:{function(**grid_dict)}")
        logger.debug(f"generation reshape:{self.generation}")
        logger.debug(f"boundary_condition_array:{self.boundary_condition_array}")
        self.set_boundary_condition_array()
